import os
import os.path as osp
import tempfile
from collections import OrderedDict
from typing import List, Optional

import pandas as pd
import tensorflow as tf
from sacred.observers import FileStorageObserver

import imitation.util.sacred as sacred_util
from imitation.scripts.config.analyze import analysis_ex
from imitation.util.sacred import dict_get_nested as get


@analysis_ex.command
def gather_tb_directories(
    source_dir: str,
    run_name: Optional[str],
    env_name: Optional[str],
    skip_failed_runs: bool,
) -> dict:
    """Gather Tensorboard directories from a `parallel_ex` run.

    The directories are copied to a unique directory in `/tmp/analysis_tb/` under
    subdirectories matching the Tensorboard events' Ray Tune trial names.

    Undocumented arguments are the same as in `analyze_imitation()`.

    Args:
      source_dir: A local_dir for Ray. For example, `~/ray_results/`.

    Returns:
      A dict with two keys. "gather_dir" (str) is a path to a /tmp/
      directory containing all the TensorBoard runs filtered from `source_dir`.
      "n_tb_dirs" (int) is the number of TensorBoard directories that were
      filtered.
    """
    sacred_dicts = _get_sacred_dicts(source_dir, run_name, env_name, skip_failed_runs)
    os.makedirs("/tmp/analysis_tb", exist_ok=True)
    tmp_dir = tempfile.mkdtemp(dir="/tmp/analysis_tb/")

    tb_dirs_count = 0
    for sd in sacred_dicts:
        # Expecting a path like "~/ray_results/{run_name}/sacred/1".
        # Want to search for all Tensorboard dirs inside
        # "~/ray_results/{run_name}".
        sacred_dir = sd.sacred_dir.rstrip("/")
        run_dir = osp.dirname(osp.dirname(sacred_dir))
        run_name = osp.basename(run_dir)

        # "tb" is TensorBoard directory built by our codebase. "sb_tb" is Stable
        # Baselines TensorBoard directory. There should be at most one of each
        # directory.
        for basename in ["tb", "sb_tb"]:
            tb_src_dirs = tuple(
                sacred_util.filter_subdirs(
                    run_dir, lambda path: osp.basename(path) == basename
                )
            )
            if tb_src_dirs:
                assert len(tb_src_dirs) == 1, "expect at most one TB dir of each type"
                tb_src_dir = tb_src_dirs[0]

                symlinks_dir = osp.join(tmp_dir, basename)
                os.makedirs(symlinks_dir, exist_ok=True)

                tb_symlink = osp.join(symlinks_dir, run_name)
                os.symlink(tb_src_dir, tb_symlink)
                tb_dirs_count += 1

    tf.logging.info(f"Symlinked {tb_dirs_count} TensorBoard dirs to {tmp_dir}.")
    tf.logging.info(f"Start Tensorboard with `tensorboard --logdir {tmp_dir}`.")
    return {"n_tb_dirs": tb_dirs_count, "gather_dir": tmp_dir}


@analysis_ex.command
def analyze_imitation(
    source_dir: str,
    run_name: Optional[str],
    env_name: Optional[str],
    skip_failed_runs: bool,
    csv_output_path: Optional[str],
    verbose: bool,
) -> pd.DataFrame:
    """Parse Sacred logs and generate a DataFrame for imitation learning results.

    Args:
      source_dir: A directory containing Sacred FileObserver subdirectories
        associated with the `train_adversarial` Sacred script. Behavior is
        undefined if there are Sacred subdirectories associated with other
        scripts.
      run_name: If provided, then only analyze results from Sacred directories
        associated with this run name. `run_name` is compared against the
        "experiment.name" key in `run.json`.
      skip_failed_runs: If True, then filter out runs where the status is FAILED.
      csv_output_path: If provided, then save a CSV output file to this path.
      verbose: If True, then print the dataframe.

    Returns:
      A list of dictionaries used to generate the analysis DataFrame.
    """
    sacred_dicts = _get_sacred_dicts(source_dir, run_name, env_name, skip_failed_runs)

    rows = []
    for sd in sacred_dicts:
        row = OrderedDict()
        rows.append(row)

        # Use get to prevent exceptions when reading in-progress experiments.
        row["status"] = get(sd.run, "status")
        row["use_gail"] = get(sd.config, "init_trainer_kwargs.use_gail")
        row["env_name"] = get(sd.config, "env_name")
        row["n_expert_demos"] = get(sd.config, "n_expert_demos")
        row["run_name"] = get(sd.run, "experiment.name")

        imit_stats = get(sd.run, "result.imit_stats")
        expert_stats = get(sd.run, "result.expert_stats")
        if imit_stats is not None and expert_stats is not None:
            # Assume that `result.imit_stats` and `result.expert_stats` are
            # formatted correctly.
            row["expert_return_summary"] = _make_return_summary(expert_stats)
            row["imit_return_summary"] = _make_return_summary(imit_stats, "monitor_")
            row["imit_expert_ratio"] = (
                imit_stats["monitor_return_mean"] / expert_stats["return_mean"]
            )

    df = pd.DataFrame(rows)
    if csv_output_path is not None:
        df.to_csv(csv_output_path)
    if verbose:
        print(df.to_string())
    return df


def _make_return_summary(stats: dict, prefix="") -> str:
    return "{:3g} ± {:3g} (n={})".format(
        stats[f"{prefix}return_mean"], stats[f"{prefix}return_std"], stats["n_traj"]
    )


def _get_sacred_dicts(
    source_dir: str, run_name: str, env_name: str, skip_failed_runs: bool
) -> List[sacred_util.SacredDicts]:
    sacred_dirs = sacred_util.filter_subdirs(source_dir)
    sacred_dicts = [
        sacred_util.SacredDicts.load_from_dir(sacred_dir) for sacred_dir in sacred_dirs
    ]

    if run_name is not None:
        sacred_dicts = filter(
            lambda sd: get(sd.run, "experiment.name") == run_name, sacred_dicts
        )

    if env_name is not None:
        sacred_dicts = filter(
            lambda sd: get(sd.config, "env_name") == env_name, sacred_dicts
        )

    if skip_failed_runs:
        sacred_dicts = filter(
            lambda sd: get(sd.run, "status") != "FAILED", sacred_dicts
        )

    return list(sacred_dicts)


def main_console():
    observer = FileStorageObserver(osp.join("output", "sacred", "analyze"))
    analysis_ex.observers.append(observer)
    analysis_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



DEFAULT_INIT_RL_KWARGS = dict(
    # For recommended PPO hyperparams in each environment, see:
    # https://github.com/araffin/rl-baselines-zoo/blob/master/hyperparams/ppo2.yml
    learning_rate=3e-4,
    nminibatches=32,
    noptepochs=10,
    ent_coef=0.0,
)



"""Behavioural Cloning (BC).

Trains policy by applying supervised learning to a fixed dataset of (observation,
action) pairs generated by some expert demonstrator.
"""

import collections
import os
from typing import Any, Callable, List, Mapping, Optional, Type, Union

import cloudpickle
import gym
import numpy as np
import tensorflow as tf
from stable_baselines import logger
from stable_baselines.common.policies import ActorCriticPolicy, BasePolicy
from tqdm.autonotebook import trange

from imitation.data import datasets, types
from imitation.policies.base import FeedForward32Policy


def set_tf_vars(
    *,
    values: List[np.ndarray],
    scope: Optional[str] = None,
    tf_vars: Optional[List[tf.Variable]] = None,
    sess: Optional[tf.Session] = None,
):
    """Set a collection of variables to take the values in `values`.

    Variables can be either specified by scope or passed directly into the
    function as a list. Variables and values will be matched based on the order
    in which they appear in their respective collections, so there must be as
    many values as variables.

    Args:
        values: list of values to load into variables.
        scope: scope to collect variables from. Either this argument xor
          `tf_vars` must be given.
        tf_vars: explicit list of TF variables to write to. Mutex with `scope`.
        sess: TF session to use, if not the default.
    """
    if scope is not None:
        assert tf_vars is None, "must give either `tf_vars` xor `scope` kwargs"
        tf_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=scope)
    else:
        assert tf_vars is not None, "must give either `tf_vars` xor `scope` kwargs"
    assert len(tf_vars) == len(values), (
        f"{len(tf_vars)} tf variables but " f"{len(values)} values supplied"
    )
    sess = sess or tf.get_default_session()
    assert sess is not None, "must supply session or have one in context"
    placeholders = [tf.placeholder(shape=v.shape, dtype=v.dtype) for v in tf_vars]
    assign_ops = [tf.assign(var, ph) for var, ph in zip(tf_vars, placeholders)]
    sess.run(
        assign_ops, feed_dict={ph: value for ph, value in zip(placeholders, values)}
    )


class BC:
    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        *,
        policy_class: Type[ActorCriticPolicy] = FeedForward32Policy,
        policy_kwargs: Optional[Mapping[str, Any]] = None,
        expert_data: Union[
            types.TransitionsMinimal,
            datasets.Dataset[types.TransitionsMinimal],
            None,
        ] = None,
        batch_size: int = 32,
        optimizer_cls: Type[tf.train.Optimizer] = tf.train.AdamOptimizer,
        optimizer_kwargs: Optional[dict] = None,
        ent_weight: float = 1e-3,
        l2_weight: float = 0.0,
    ):
        """Behavioral cloning (BC).

        Recovers a policy via supervised learning on a Dataset of observation-action
        pairs.

        Args:
            env: environment to train on.
            policy_class: used to instantiate imitation policy.
            expert_data: If not None, then immediately call
                  `self.set_expert_dataset(expert_data)` during initialization.
            batch_size: batch size used for training.
            optimizer_cls: optimiser to use for supervised training.
            optimizer_kwargs: keyword arguments for optimiser construction.
        """
        self.action_space = action_space
        self.observation_space = observation_space
        self.policy_class = policy_class
        policy_kwargs_defaults = dict(
            ob_space=self.observation_space,
            ac_space=self.action_space,
            n_batch=None,
            n_env=1,
            n_steps=1000,
        )
        self.policy_kwargs = collections.ChainMap(
            policy_kwargs or {}, policy_kwargs_defaults
        )

        assert batch_size >= 1
        self.batch_size = batch_size
        self.expert_dataset: Optional[datasets.Dataset[types.TransitionsMinimal]] = None
        self.sess = tf.get_default_session()
        assert self.sess is not None, "need to construct this within a session scope"
        self.ent_weight = ent_weight
        self.l2_weight = l2_weight

        self.optimizer_cls = optimizer_cls
        self.optimizer_kwargs = optimizer_kwargs or {}

        self._build_tf_graph()
        self.sess.run(tf.global_variables_initializer())
        if expert_data is not None:
            self.set_expert_dataset(expert_data)

    def set_expert_dataset(
        self,
        expert_data: Union[
            types.TransitionsMinimal,
            datasets.Dataset[types.TransitionsMinimal],
        ],
    ):
        """Replace the current expert dataset with a new one.

        Useful for DAgger and other interactive algorithms.

        Args:
             expert_data: Either a `Dataset[types.TransitionsMinimal]` for which
                 `.size()` is not None, or a instance of `TransitionsMinimal`, which
                 is automatically converted to a shuffled, epoch-order
                 `Dataset[types.TransitionsMinimal]`.
        """
        if isinstance(expert_data, types.TransitionsMinimal):
            trans = expert_data
            expert_dataset = datasets.TransitionsDictDatasetAdaptor(
                trans, datasets.EpochOrderDictDataset
            )
        else:
            assert isinstance(expert_data, datasets.Dataset)
            expert_dataset = expert_data
        assert expert_dataset.size() is not None
        self.expert_dataset = expert_dataset

    def _build_tf_graph(self):
        with tf.variable_scope("bc_supervised_loss"):
            with tf.variable_scope("model") as model_scope:
                self.policy = self.policy_class(
                    sess=self.sess, **self.policy_kwargs
                )  # pytype: disable=not-instantiable
                inner_scope = tf.get_variable_scope().name
                self.policy_variables = tf.get_collection(
                    tf.GraphKeys.TRAINABLE_VARIABLES, scope=inner_scope
                )
            self._true_acts_ph = self.policy.pdtype.sample_placeholder(
                [None], name="ref_acts_ph"
            )
            self._entropy = tf.reduce_mean(self.policy.proba_distribution.entropy())
            self._ent_loss = -self.ent_weight * self._entropy
            self._neglogp = tf.reduce_mean(
                self.policy.proba_distribution.neglogp(self._true_acts_ph)
            )

            _l2_norms = [tf.nn.l2_loss(w) for w in model_scope.trainable_variables()]
            self._l2_norm = tf.reduce_sum(_l2_norms)
            self._l2_loss = self._l2_norm * self.l2_weight

            self._loss = self._neglogp + self._ent_loss + self._l2_loss
            opt = self.optimizer_cls(**self.optimizer_kwargs)
            self._train_op = opt.minimize(self._loss)

            self._prob_true_act = tf.reduce_mean(
                tf.exp(-self.policy.proba_distribution.neglogp(self._true_acts_ph))
            )
            self._stats_dict = dict(
                neglopp=self._neglogp,
                loss=self._loss,
                entropy=self._entropy,
                ent_loss=self._ent_loss,
                prob_true_act=self._prob_true_act,
                l2_norm=self._l2_norm,
                l2_loss=self._l2_loss,
            )

    def train(
        self,
        n_epochs: int = 100,
        *,
        on_epoch_end: Callable[[dict], None] = None,
        log_interval: int = 100,
    ):
        """Train with supervised learning for some number of epochs.

        Here an 'epoch' is just a complete pass through the expert transition
        dataset.

        Args:
          n_epochs: number of complete passes made through dataset.
          on_epoch_end: optional callback to run at
            the end of each epoch. Will receive all locals from this function as
            dictionary argument (!!).
        """
        assert self.batch_size >= 1
        samples_so_far = 0
        batch_num = 0
        for epoch_num in trange(n_epochs, desc="BC epoch"):
            while samples_so_far < (epoch_num + 1) * self.expert_dataset.size():
                batch_num += 1
                trans = self.expert_dataset.sample(self.batch_size)
                assert len(trans) == self.batch_size
                samples_so_far += self.batch_size
                feed_dict = {
                    self._true_acts_ph: trans.acts,
                    self.policy.obs_ph: trans.obs,
                }
                _, stats_dict = self.sess.run(
                    [self._train_op, self._stats_dict], feed_dict=feed_dict
                )
                stats_dict["epoch_num"] = epoch_num
                stats_dict["n_updates"] = batch_num
                stats_dict["batch_size"] = self.batch_size

                if batch_num % log_interval == 0:
                    for k, v in stats_dict.items():
                        logger.logkv(k, v)
                    logger.dumpkvs()
                batch_num += 1

            if on_epoch_end is not None:
                on_epoch_end(locals())

    def save_policy(self, policy_path: str):
        """Save a policy to a pickle. Can be reloaded by `.reconstruct_policy()`.

        Args:
            policy_path: path to save policy to.
        """
        policy_params = self.sess.run(self.policy_variables)
        data = {
            "class": self.policy_class,
            "kwargs": self.policy_kwargs,
            "params": policy_params,
        }
        dirname = os.path.dirname(policy_path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        with open(policy_path, "wb") as fp:
            cloudpickle.dump(data, fp)

    @staticmethod
    def reconstruct_policy(
        policy_path: str,
        sess: Optional[tf.Session] = None,
    ) -> BasePolicy:
        """Reconstruct a saved policy.

        Args:
            policy_path: path a policy produced by `.save_policy()`.
            sess: optional session to construct policy under,
              if not the default session.

        Returns:
            policy: policy with reloaded weights.
        """
        if sess is None:
            sess = tf.get_default_session()
            assert sess is not None, "must supply session via kwarg or context mgr"

        # re-read data from dict
        with open(policy_path, "rb") as fp:
            loaded_pickle = cloudpickle.load(fp)

        # construct the policy class
        klass = loaded_pickle["class"]
        kwargs = loaded_pickle["kwargs"]
        with tf.variable_scope("reconstructed_policy"):
            rv_pol = klass(sess=sess, **kwargs)
            inner_scope = tf.get_variable_scope().name

        # set values for the new policy's parameters
        param_values = loaded_pickle["params"]
        set_tf_vars(values=param_values, scope=inner_scope, sess=sess)

        return rv_pol



import numpy as np

from imitation.envs.examples.airl_envs.dynamic_mjc.model_builder import MJCModel


def block_push(object_pos=(0, 0, 0), goal_pos=(0, 0, 0)):
    mjcmodel = MJCModel("block_push")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="radian", coordinate="local")
    mjcmodel.root.option(
        timestep="0.01", gravity="0 0 0", iterations="20", integrator="Euler"
    )
    default = mjcmodel.root.default()
    default.joint(armature="0.04", damping=1, limited="true")
    default.geom(
        friction=".8 .1 .1",
        density="300",
        margin="0.002",
        condim="1",
        contype="1",
        conaffinity="1",
    )

    worldbody = mjcmodel.root.worldbody()

    palm = worldbody.body(name="palm", pos=[0, 0, 0])
    palm.geom(
        name="palm_geom", type="capsule", fromto=[0, -0.1, 0, 0, 0.1, 0], size=0.12
    )
    proximal1 = palm.body(name="proximal_1", pos=[0, 0, 0])
    proximal1.joint(
        name="proximal_j_1",
        type="hinge",
        pos=[0, 0, 0],
        axis=[0, 1, 0],
        range=[-2.5, 2.3],
    )
    proximal1.geom(
        type="capsule", fromto=[0, 0, 0, 0.4, 0, 0], size=0.06, contype=1, conaffinity=1
    )
    distal1 = proximal1.body(name="distal_1", pos=[0.4, 0, 0])
    distal1.joint(
        name="distal_j_1",
        type="hinge",
        pos="0 0 0",
        axis="0 1 0",
        range="-2.3213 2.3",
        damping="1.0",
    )
    distal1.geom(
        type="capsule",
        fromto="0 0 0 0.4 0 0",
        size="0.06",
        contype="1",
        conaffinity="1",
    )
    distal2 = distal1.body(name="distal_2", pos=[0.4, 0, 0])
    distal2.joint(
        name="distal_j_2",
        type="hinge",
        pos="0 0 0",
        axis="0 1 0",
        range="-2.3213 2.3",
        damping="1.0",
    )
    distal2.geom(
        type="capsule",
        fromto="0 0 0 0.4 0 0",
        size="0.06",
        contype="1",
        conaffinity="1",
    )
    distal4 = distal2.body(name="distal_4", pos=[0.4, 0, 0])
    distal4.site(name="tip arml", pos="0.1 0 -0.2", size="0.01")
    distal4.site(name="tip armr", pos="0.1 0 0.2", size="0.01")
    distal4.joint(
        name="distal_j_3",
        type="hinge",
        pos="0 0 0",
        axis="1 0 0",
        range="-3.3213 3.3",
        damping="0.5",
    )
    distal4.geom(
        type="capsule",
        fromto="0 0 -0.2 0 0 0.2",
        size="0.04",
        contype="1",
        conaffinity="1",
    )
    distal4.geom(
        type="capsule",
        fromto="0 0 -0.2 0.2 0 -0.2",
        size="0.04",
        contype="1",
        conaffinity="1",
    )
    distal4.geom(
        type="capsule",
        fromto="0 0 0.2 0.2 0 0.2",
        size="0.04",
        contype="1",
        conaffinity="1",
    )

    obj = worldbody.body(name="obj", pos=object_pos)
    obj.geom(
        rgba="1. 1. 1. 1",
        type="box",
        size="0.05 0.05 0.05",
        density="0.00001",
        contype="1",
        conaffinity="1",
    )
    obj.joint(
        name="obj_slidez",
        type="slide",
        pos="0.025 0.025 0.025",
        axis="0 0 1",
        range="-10.3213 10.3",
        damping="0.5",
    )
    obj.joint(
        name="obj_slidex",
        type="slide",
        pos="0.025 0.025 0.025",
        axis="1 0 0",
        range="-10.3213 10.3",
        damping="0.5",
    )
    distal10 = obj.body(name="distal_10", pos=[0, 0, 0])
    distal10.site(name="obj_pos", pos=[0.025, 0.025, 0.025], size=0.01)

    goal = worldbody.body(name="goal", pos=goal_pos)
    goal.geom(
        rgba="1. 0. 0. 1",
        type="box",
        size="0.1 0.1 0.1",
        density="0.00001",
        contype="0",
        conaffinity="0",
    )
    distal11 = goal.body(name="distal_11", pos=[0, 0, 0])
    distal11.site(name="goal_pos", pos=[0.05, 0.05, 0.05], size=0.01)

    actuator = mjcmodel.root.actuator()
    actuator.motor(joint="proximal_j_1", ctrlrange="-2 2", ctrllimited="true")
    actuator.motor(joint="distal_j_1", ctrlrange="-2 2", ctrllimited="true")
    actuator.motor(joint="distal_j_2", ctrlrange="-2 2", ctrllimited="true")
    actuator.motor(joint="distal_j_3", ctrlrange="-2 2", ctrllimited="true")

    return mjcmodel


EAST = 0
WEST = 1
NORTH = 2
SOUTH = 3


def twod_corridor(direction=EAST, length=1.2):
    mjcmodel = MJCModel("twod_corridor")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="radian", coordinate="local")
    mjcmodel.root.option(
        timestep="0.01", gravity="0 0 0", iterations="20", integrator="Euler"
    )
    default = mjcmodel.root.default()
    default.joint(damping=1, limited="false")
    default.geom(
        friction=".5 .1 .1",
        density="1000",
        margin="0.002",
        condim="1",
        contype="2",
        conaffinity="1",
    )

    worldbody = mjcmodel.root.worldbody()

    particle = worldbody.body(name="particle", pos=[0, 0, 0])
    particle.geom(
        name="particle_geom",
        type="sphere",
        size="0.03",
        rgba="0.0 0.0 1.0 1",
        contype=1,
    )
    particle.site(name="particle_site", pos=[0, 0, 0], size=0.01)
    particle.joint(name="ball_x", type="slide", pos=[0, 0, 0], axis=[1, 0, 0])
    particle.joint(name="ball_y", type="slide", pos=[0, 0, 0], axis=[0, 1, 0])

    pos = np.array([0.0, 0, 0])
    if direction == EAST or direction == WEST:
        pos[0] = length - 0.1
    else:
        pos[1] = length - 0.1
    if direction == WEST or direction == SOUTH:
        pos = -pos

    target = worldbody.body(name="target", pos=pos)
    target.geom(
        name="target_geom",
        conaffinity=2,
        type="sphere",
        size=0.02,
        rgba=[0, 0.9, 0.1, 1],
    )

    # arena
    if direction == EAST:
        L = -0.1
        R = length
        U = 0.1
        D = -0.1
    elif direction == WEST:
        L = -length
        R = 0.1
        U = 0.1
        D = -0.1
    elif direction == SOUTH:
        L = -0.1
        R = 0.1
        U = 0.1
        D = -length
    elif direction == NORTH:
        L = -0.1
        R = 0.1
        U = length
        D = -0.1

    worldbody.geom(
        conaffinity=1,
        fromto=[L, D, 0.01, R, D, 0.01],
        name="sideS",
        rgba="0.9 0.4 0.6 1",
        size=0.02,
        type="capsule",
    )
    worldbody.geom(
        conaffinity=1,
        fromto=[R, D, 0.01, R, U, 0.01],
        name="sideE",
        rgba="0.9 0.4 0.6 1",
        size=".02",
        type="capsule",
    )
    worldbody.geom(
        conaffinity=1,
        fromto=[L, U, 0.01, R, U, 0.01],
        name="sideN",
        rgba="0.9 0.4 0.6 1",
        size=".02",
        type="capsule",
    )
    worldbody.geom(
        conaffinity=1,
        fromto=[L, D, 0.01, L, U, 0.01],
        name="sideW",
        rgba="0.9 0.4 0.6 1",
        size=".02",
        type="capsule",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(joint="ball_x", ctrlrange=[-1.0, 1.0], ctrllimited=True)
    actuator.motor(joint="ball_y", ctrlrange=[-1.0, 1.0], ctrllimited=True)

    return mjcmodel


LEFT = 0
RIGHT = 1


def point_mass_maze(direction=RIGHT, length=1.2, borders=True):
    mjcmodel = MJCModel("twod_maze")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="radian", coordinate="local")
    mjcmodel.root.option(
        timestep="0.01", gravity="0 0 0", iterations="20", integrator="Euler"
    )
    default = mjcmodel.root.default()
    default.joint(damping=1, limited="false")
    default.geom(
        friction=".5 .1 .1",
        density="1000",
        margin="0.002",
        condim="1",
        contype="2",
        conaffinity="1",
    )

    worldbody = mjcmodel.root.worldbody()

    particle = worldbody.body(name="particle", pos=[length / 2, 0, 0])
    particle.geom(
        name="particle_geom",
        type="sphere",
        size="0.03",
        rgba="0.0 0.0 1.0 1",
        contype=1,
    )
    particle.site(name="particle_site", pos=[0, 0, 0], size=0.01)
    particle.joint(name="ball_x", type="slide", pos=[0, 0, 0], axis=[1, 0, 0])
    particle.joint(name="ball_y", type="slide", pos=[0, 0, 0], axis=[0, 1, 0])

    target = worldbody.body(name="target", pos=[length / 2, length - 0.1, 0])
    target.geom(
        name="target_geom",
        conaffinity=2,
        type="sphere",
        size=0.02,
        rgba=[0, 0.9, 0.1, 1],
    )

    L = -0.1
    R = length
    U = length
    D = -0.1

    if borders:
        worldbody.geom(
            conaffinity=1,
            fromto=[L, D, 0.01, R, D, 0.01],
            name="sideS",
            rgba="0.9 0.4 0.6 1",
            size=".02",
            type="capsule",
        )
        worldbody.geom(
            conaffinity=1,
            fromto=[R, D, 0.01, R, U, 0.01],
            name="sideE",
            rgba="0.9 0.4 0.6 1",
            size=".02",
            type="capsule",
        )
        worldbody.geom(
            conaffinity=1,
            fromto=[L, U, 0.01, R, U, 0.01],
            name="sideN",
            rgba="0.9 0.4 0.6 1",
            size=".02",
            type="capsule",
        )
        worldbody.geom(
            conaffinity=1,
            fromto=[L, D, 0.01, L, U, 0.01],
            name="sideW",
            rgba="0.9 0.4 0.6 1",
            size=".02",
            type="capsule",
        )

    # arena
    if direction == LEFT:
        BL = -0.1
        BR = length * 2 / 3
        BH = length / 2
    else:
        BL = length * 1 / 3
        BR = length
        BH = length / 2

    worldbody.geom(
        conaffinity=1,
        fromto=[BL, BH, 0.01, BR, BH, 0.01],
        name="barrier",
        rgba="0.9 0.4 0.6 1",
        size=".02",
        type="capsule",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(joint="ball_x", ctrlrange=[-1.0, 1.0], ctrllimited=True)
    actuator.motor(joint="ball_y", ctrlrange=[-1.0, 1.0], ctrllimited=True)

    return mjcmodel


def ant_maze(direction=RIGHT, length=6.0):
    mjcmodel = MJCModel("ant_maze")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="degree", coordinate="local")
    mjcmodel.root.option(
        timestep="0.01", gravity="0 0 -9.8", iterations="20", integrator="Euler"
    )

    assets = mjcmodel.root.asset()
    assets.texture(
        builtin="gradient",
        height="100",
        rgb1="1 1 1",
        rgb2="0 0 0",
        type="skybox",
        width="100",
    )
    assets.texture(
        builtin="flat",
        height="1278",
        mark="cross",
        markrgb="1 1 1",
        name="texgeom",
        random="0.01",
        rgb1="0.8 0.6 0.4",
        rgb2="0.8 0.6 0.4",
        type="cube",
        width="127",
    )
    assets.texture(
        builtin="checker",
        height="100",
        name="texplane",
        rgb1="0 0 0",
        rgb2="0.8 0.8 0.8",
        type="2d",
        width="100",
    )
    assets.material(
        name="MatPlane",
        reflectance="0.5",
        shininess="1",
        specular="1",
        texrepeat="60 60",
        texture="texplane",
    )
    assets.material(name="geom", texture="texgeom", texuniform="true")

    default = mjcmodel.root.default()
    default.joint(armature="1", damping=1, limited="true")
    default.geom(
        friction="1 0.5 0.5", density="5.0", margin="0.01", condim="3", conaffinity="0"
    )

    worldbody = mjcmodel.root.worldbody()

    ant = worldbody.body(name="ant", pos=[length / 2, 1.0, 0.05])
    ant.geom(name="torso_geom", pos=[0, 0, 0], size="0.25", type="sphere")
    ant.joint(
        armature="0",
        damping="0",
        limited="false",
        margin="0.01",
        name="root",
        pos=[0, 0, 0],
        type="free",
    )

    front_left_leg = ant.body(name="front_left_leg", pos=[0, 0, 0])
    front_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="aux_1_geom",
        size="0.08",
        type="capsule",
    )
    aux_1 = front_left_leg.body(name="aux_1", pos=[0.2, 0.2, 0])
    aux_1.joint(
        axis=[0, 0, 1], name="hip_1", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="left_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_1 = aux_1.body(pos=[0.2, 0.2, 0])
    ankle_1.joint(
        axis=[-1, 1, 0],
        name="ankle_1",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, 0.4, 0.0],
        name="left_ankle_geom",
        size="0.08",
        type="capsule",
    )

    front_right_leg = ant.body(name="front_right_leg", pos=[0, 0, 0])
    front_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="aux_2_geom",
        size="0.08",
        type="capsule",
    )
    aux_2 = front_right_leg.body(name="aux_2", pos=[-0.2, 0.2, 0])
    aux_2.joint(
        axis=[0, 0, 1], name="hip_2", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="right_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_2 = aux_2.body(pos=[-0.2, 0.2, 0])
    ankle_2.joint(
        axis=[1, 1, 0],
        name="ankle_2",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, 0.4, 0.0],
        name="right_ankle_geom",
        size="0.08",
        type="capsule",
    )

    back_left_leg = ant.body(name="back_left_leg", pos=[0, 0, 0])
    back_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="aux_3_geom",
        size="0.08",
        type="capsule",
    )
    aux_3 = back_left_leg.body(name="aux_3", pos=[-0.2, -0.2, 0])
    aux_3.joint(
        axis=[0, 0, 1], name="hip_3", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_3.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="backleft_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_3 = aux_3.body(pos=[-0.2, -0.2, 0])
    ankle_3.joint(
        axis=[-1, 1, 0],
        name="ankle_3",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_3.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, -0.4, 0.0],
        name="backleft_ankle_geom",
        size="0.08",
        type="capsule",
    )

    back_right_leg = ant.body(name="back_right_leg", pos=[0, 0, 0])
    back_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="aux_4_geom",
        size="0.08",
        type="capsule",
    )
    aux_4 = back_right_leg.body(name="aux_4", pos=[0.2, -0.2, 0])
    aux_4.joint(
        axis=[0, 0, 1], name="hip_4", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_4.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="backright_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_4 = aux_4.body(pos=[0.2, -0.2, 0])
    ankle_4.joint(
        axis=[1, 1, 0],
        name="ankle_4",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_4.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, -0.4, 0.0],
        name="backright_ankle_geom",
        size="0.08",
        type="capsule",
    )

    target = worldbody.body(name="target", pos=[length / 2, length - 0.2, -0.5])
    target.geom(
        name="target_geom",
        conaffinity=2,
        type="sphere",
        size=0.2,
        rgba=[0, 0.9, 0.1, 1],
    )

    L = length / 2
    h = 0.75
    w = 0.05

    worldbody.geom(
        conaffinity=1,
        name="sideS",
        rgba="0.9 0.4 0.6 1",
        size=[L, w, h],
        pos=[length / 2, 0, 0],
        type="box",
    )
    worldbody.geom(
        conaffinity=1,
        name="sideE",
        rgba="0.9 0.4 0.6 1",
        size=[w, L, h],
        pos=[length, length / 2, 0],
        type="box",
    )
    worldbody.geom(
        conaffinity=1,
        name="sideN",
        rgba="0.9 0.4 0.6 1",
        size=[L, w, h],
        pos=[length / 2, length, 0],
        type="box",
    )
    worldbody.geom(
        conaffinity=1,
        name="sideW",
        rgba="0.9 0.4 0.6 1",
        size=[w, L, h],
        pos=[0, length / 2, 0],
        type="box",
    )

    # arena
    if direction == LEFT:
        bx, by, bz = (length / 3, length / 2, 0)
    else:
        bx, by, bz = (length * 2 / 3, length / 2, 0)

    worldbody.geom(
        conaffinity=1,
        name="barrier",
        rgba="0.9 0.4 0.6 1",
        size=[L * 2 / 3, w, h],
        pos=[bx, by, bz],
        type="box",
    )
    worldbody.geom(
        conaffinity="1",
        condim="3",
        material="MatPlane",
        name="floor",
        pos=[length / 2, length / 2, -h + w],
        rgba="0.8 0.9 0.8 1",
        size="40 40 40",
        type="plane",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_4", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_4", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_1", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_1", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_2", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_2", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_3", gear="50")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_3", gear="50")

    return mjcmodel


def ant_maze_corridor(direction=RIGHT, height=6.0, width=10.0):
    mjcmodel = MJCModel("ant_maze_corridor")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="degree", coordinate="local")
    mjcmodel.root.option(
        timestep="0.01", gravity="0 0 -9.8", iterations="20", integrator="Euler"
    )

    assets = mjcmodel.root.asset()
    assets.texture(
        builtin="gradient",
        height="100",
        rgb1="1 1 1",
        rgb2="0 0 0",
        type="skybox",
        width="100",
    )
    assets.texture(
        builtin="flat",
        height="1278",
        mark="cross",
        markrgb="1 1 1",
        name="texgeom",
        random="0.01",
        rgb1="0.8 0.6 0.4",
        rgb2="0.8 0.6 0.4",
        type="cube",
        width="127",
    )
    assets.texture(
        builtin="checker",
        height="100",
        name="texplane",
        rgb1="0 0 0",
        rgb2="0.8 0.8 0.8",
        type="2d",
        width="100",
    )
    assets.material(
        name="MatPlane",
        reflectance="0.5",
        shininess="1",
        specular="1",
        texrepeat="60 60",
        texture="texplane",
    )
    assets.material(name="geom", texture="texgeom", texuniform="true")

    default = mjcmodel.root.default()
    default.joint(armature="1", damping=1, limited="true")
    default.geom(
        friction="1 0.5 0.5", density="5.0", margin="0.01", condim="3", conaffinity="0"
    )

    worldbody = mjcmodel.root.worldbody()

    ant = worldbody.body(name="ant", pos=[height / 2, 1.0, 0.05])
    ant.geom(name="torso_geom", pos=[0, 0, 0], size="0.25", type="sphere")
    ant.joint(
        armature="0",
        damping="0",
        limited="false",
        margin="0.01",
        name="root",
        pos=[0, 0, 0],
        type="free",
    )

    front_left_leg = ant.body(name="front_left_leg", pos=[0, 0, 0])
    front_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="aux_1_geom",
        size="0.08",
        type="capsule",
    )
    aux_1 = front_left_leg.body(name="aux_1", pos=[0.2, 0.2, 0])
    aux_1.joint(
        axis=[0, 0, 1], name="hip_1", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, 0.2, 0.0],
        name="left_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_1 = aux_1.body(pos=[0.2, 0.2, 0])
    ankle_1.joint(
        axis=[-1, 1, 0],
        name="ankle_1",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_1.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, 0.4, 0.0],
        name="left_ankle_geom",
        size="0.08",
        type="capsule",
    )

    front_right_leg = ant.body(name="front_right_leg", pos=[0, 0, 0])
    front_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="aux_2_geom",
        size="0.08",
        type="capsule",
    )
    aux_2 = front_right_leg.body(name="aux_2", pos=[-0.2, 0.2, 0])
    aux_2.joint(
        axis=[0, 0, 1], name="hip_2", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, 0.2, 0.0],
        name="right_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_2 = aux_2.body(pos=[-0.2, 0.2, 0])
    ankle_2.joint(
        axis=[1, 1, 0],
        name="ankle_2",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_2.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, 0.4, 0.0],
        name="right_ankle_geom",
        size="0.08",
        type="capsule",
    )

    back_left_leg = ant.body(name="back_left_leg", pos=[0, 0, 0])
    back_left_leg.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="aux_3_geom",
        size="0.08",
        type="capsule",
    )
    aux_3 = back_left_leg.body(name="aux_3", pos=[-0.2, -0.2, 0])
    aux_3.joint(
        axis=[0, 0, 1], name="hip_3", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_3.geom(
        fromto=[0.0, 0.0, 0.0, -0.2, -0.2, 0.0],
        name="backleft_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_3 = aux_3.body(pos=[-0.2, -0.2, 0])
    ankle_3.joint(
        axis=[-1, 1, 0],
        name="ankle_3",
        pos=[0.0, 0.0, 0.0],
        range=[-70, -30],
        type="hinge",
    )
    ankle_3.geom(
        fromto=[0.0, 0.0, 0.0, -0.4, -0.4, 0.0],
        name="backleft_ankle_geom",
        size="0.08",
        type="capsule",
    )

    back_right_leg = ant.body(name="back_right_leg", pos=[0, 0, 0])
    back_right_leg.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="aux_4_geom",
        size="0.08",
        type="capsule",
    )
    aux_4 = back_right_leg.body(name="aux_4", pos=[0.2, -0.2, 0])
    aux_4.joint(
        axis=[0, 0, 1], name="hip_4", pos=[0.0, 0.0, 0.0], range=[-30, 30], type="hinge"
    )
    aux_4.geom(
        fromto=[0.0, 0.0, 0.0, 0.2, -0.2, 0.0],
        name="backright_leg_geom",
        size="0.08",
        type="capsule",
    )
    ankle_4 = aux_4.body(pos=[0.2, -0.2, 0])
    ankle_4.joint(
        axis=[1, 1, 0],
        name="ankle_4",
        pos=[0.0, 0.0, 0.0],
        range=[30, 70],
        type="hinge",
    )
    ankle_4.geom(
        fromto=[0.0, 0.0, 0.0, 0.4, -0.4, 0.0],
        name="backright_ankle_geom",
        size="0.08",
        type="capsule",
    )

    target = worldbody.body(name="target", pos=[height / 2, width - 1.0, -0.5])
    target.geom(
        name="target_geom",
        conaffinity=2,
        type="sphere",
        size=0.2,
        rgba=[0, 0.9, 0.1, 1],
    )

    L = height / 2
    h = 0.75
    w = 0.05

    worldbody.geom(
        conaffinity=1,
        name="sideS",
        rgba="0.9 0.4 0.6 1",
        size=[L, w, h],
        pos=[height / 2, 0, 0],
        type="box",
    )
    worldbody.geom(
        conaffinity=1,
        name="sideE",
        rgba="0.9 0.4 0.6 1",
        size=[w, width / 2, h],
        pos=[height, width / 2, 0],
        type="box",
    )
    worldbody.geom(
        conaffinity=1,
        name="sideN",
        rgba="0.9 0.4 0.6 1",
        size=[L, w, h],
        pos=[height / 2, width, 0],
        type="box",
    )
    worldbody.geom(
        conaffinity=1,
        name="sideW",
        rgba="0.9 0.4 0.6 1",
        size=[w, width / 2, h],
        pos=[0, width / 2, 0],
        type="box",
    )

    # arena
    wall_ratio = 0.55  # 2.0/3
    if direction == LEFT:
        bx, by, bz = (height * (wall_ratio / 2), width / 2, 0)
    else:
        bx, by, bz = (height * (1 - wall_ratio / 2), width / 2, 0)

    worldbody.geom(
        conaffinity=1,
        name="barrier",
        rgba="0.9 0.4 0.6 1",
        size=[L * (wall_ratio), w, h],
        pos=[bx, by, bz],
        type="box",
    )
    worldbody.geom(
        conaffinity="1",
        condim="3",
        material="MatPlane",
        name="floor",
        pos=[height / 2, height / 2, -h + w],
        rgba="0.8 0.9 0.8 1",
        size="40 40 40",
        type="plane",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_4", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_4", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_1", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_1", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_2", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_2", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="hip_3", gear="30")
    actuator.motor(ctrllimited="true", ctrlrange="-1.0 1.0", joint="ankle_3", gear="30")

    return mjcmodel


def pusher(goal_pos=np.array([0.45, -0.05, -0.323])):
    mjcmodel = MJCModel("pusher")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="radian", coordinate="local")
    mjcmodel.root.option(
        timestep="0.01", gravity="0 0 0", iterations="20", integrator="Euler"
    )
    default = mjcmodel.root.default()
    default.joint(armature=0.04, damping=1, limited=False)
    default.geom(
        friction=[0.8, 0.1, 0.1],
        density=300,
        margin=0.002,
        condim=1,
        contype=0,
        conaffinity=0,
    )

    worldbody = mjcmodel.root.worldbody()
    worldbody.light(diffuse=[0.5, 0.5, 0.5], pos=[0, 0, 3], dir=[0, 0, -1])
    worldbody.geom(
        name="table",
        type="plane",
        pos=[0, 0.5, -0.325],
        size=[1, 1, 0.1],
        contype=1,
        conaffinity=1,
    )

    r_shoulder_pan_link = worldbody.body(name="r_shoulder_pan_link", pos=[0, -0.6, 0])
    r_shoulder_pan_link.geom(
        name="e1",
        type="sphere",
        rgba=[0.6, 0.6, 0.6, 1],
        pos=[-0.06, 0.05, 0.2],
        size=0.05,
    )
    r_shoulder_pan_link.geom(
        name="e2",
        type="sphere",
        rgba=[0.6, 0.6, 0.6, 1],
        pos=[0.06, 0.05, 0.2],
        size=0.05,
    )
    r_shoulder_pan_link.geom(
        name="e1p",
        type="sphere",
        rgba=[0.1, 0.1, 0.1, 1],
        pos=[-0.06, 0.09, 0.2],
        size=0.03,
    )
    r_shoulder_pan_link.geom(
        name="e2p",
        type="sphere",
        rgba=[0.1, 0.1, 0.1, 1],
        pos=[0.06, 0.09, 0.2],
        size=0.03,
    )
    r_shoulder_pan_link.geom(
        name="sp", type="capsule", fromto=[0, 0, -0.4, 0, 0, 0.2], size=0.1
    )

    r_shoulder_pan_link.joint(
        name="r_shoulder_pan_joint",
        type="hinge",
        pos=[0, 0, 0],
        axis=[0, 0, 1],
        range=[-2.2854, 1.714602],
        damping=1.0,
    )

    r_shoulder_lift_link = r_shoulder_pan_link.body(
        name="r_shoulder_lift_link", pos=[0.1, 0, 0]
    )
    r_shoulder_lift_link.geom(
        name="s1", type="capsule", fromto="0 -0.1 0 0 0.1 0", size="0.1"
    )
    r_shoulder_lift_link.joint(
        name="r_shoulder_lift_joint",
        type="hinge",
        pos="0 0 0",
        axis="0 1 0",
        range="-0.5236 1.3963",
        damping="1.0",
    )

    r_upper_arm_roll_link = r_shoulder_lift_link.body(
        name="r_upper_arm_roll_link", pos=[0, 0, 0]
    )
    r_upper_arm_roll_link.geom(
        name="uar", type="capsule", fromto="-0.1 0 0 0.1 0 0", size="0.02"
    )
    r_upper_arm_roll_link.joint(
        name="r_upper_arm_roll_joint",
        type="hinge",
        pos="0 0 0",
        axis="1 0 0",
        range="-1.5 1.7",
        damping="0.1",
    )

    r_upper_arm_link = r_upper_arm_roll_link.body(
        name="r_upper_arm_link", pos=[0, 0, 0]
    )
    r_upper_arm_link.geom(
        name="ua", type="capsule", fromto="0 0 0 0.4 0 0", size="0.06"
    )

    r_elbow_flex_link = r_upper_arm_link.body(name="r_elbow_flex_link", pos=[0.4, 0, 0])
    r_elbow_flex_link.geom(
        name="ef", type="capsule", fromto="0 -0.02 0 0.0 0.02 0", size="0.06"
    )
    r_elbow_flex_link.joint(
        name="r_elbow_flex_joint",
        type="hinge",
        pos="0 0 0",
        axis="0 1 0",
        range="-2.3213 0",
        damping="0.1",
    )

    r_forearm_roll_link = r_elbow_flex_link.body(
        name="r_forearm_roll_link", pos=[0, 0, 0]
    )
    r_forearm_roll_link.geom(
        name="fr", type="capsule", fromto="-0.1 0 0 0.1 0 0", size="0.02"
    )
    r_forearm_roll_link.joint(
        name="r_forearm_roll_joint",
        type="hinge",
        limited="true",
        pos="0 0 0",
        axis="1 0 0",
        damping=".1",
        range="-1.5 1.5",
    )

    r_forearm_link = r_forearm_roll_link.body(name="r_forearm_link", pos=[0, 0, 0])
    r_forearm_link.geom(
        name="fa", type="capsule", fromto="0 0 0 0.291 0 0", size="0.05"
    )

    r_wrist_flex_link = r_forearm_link.body(name="r_wrist_flex_link", pos=[0.321, 0, 0])
    r_wrist_flex_link.geom(
        name="wf", type="capsule", fromto="0 -0.02 0 0 0.02 0", size="0.01"
    )
    r_wrist_flex_link.joint(
        name="r_wrist_flex_joint",
        type="hinge",
        pos="0 0 0",
        axis="0 1 0",
        range="-1.094 0",
        damping=".1",
    )

    r_wrist_roll_link = r_wrist_flex_link.body(name="r_wrist_roll_link", pos=[0, 0, 0])
    r_wrist_roll_link.joint(
        name="r_wrist_roll_joint",
        type="hinge",
        pos="0 0 0",
        limited="true",
        axis="1 0 0",
        damping="0.1",
        range="-1.5 1.5",
    )
    r_wrist_roll_link.geom(
        type="capsule",
        fromto="0 -0.1 0. 0.0 +0.1 0",
        size="0.02",
        contype="1",
        conaffinity="1",
    )
    r_wrist_roll_link.geom(
        type="capsule",
        fromto="0 -0.1 0. 0.1 -0.1 0",
        size="0.02",
        contype="1",
        conaffinity="1",
    )
    r_wrist_roll_link.geom(
        type="capsule",
        fromto="0 +0.1 0. 0.1 +0.1 0",
        size="0.02",
        contype="1",
        conaffinity="1",
    )

    tips_arm = r_wrist_roll_link.body(name="tips_arm", pos=[0, 0, 0])
    tips_arm.geom(name="tip_arml", type="sphere", pos="0.1 -0.1 0.", size="0.01")
    tips_arm.geom(name="tip_armr", type="sphere", pos="0.1 0.1 0.", size="0.01")

    object_ = worldbody.body(name="object", pos=[0.0, 0.0, -0.275])
    object_.geom(
        rgba="1 1 1 1",
        type="cylinder",
        size="0.05 0.05 0.05",
        density="0.00001",
        conaffinity="0",
        contype=1,
    )
    object_.joint(
        name="obj_slidey",
        type="slide",
        pos="0 0 0",
        axis="0 1 0",
        range="-10.3213 10.3",
        damping="0.5",
    )
    object_.joint(
        name="obj_slidex",
        type="slide",
        pos="0 0 0",
        axis="1 0 0",
        range="-10.3213 10.3",
        damping="0.5",
    )

    goal = worldbody.body(name="goal", pos=goal_pos)
    goal.geom(
        rgba="1 0 0 1",
        type="cylinder",
        size="0.08 0.001 0.1",
        density="0.00001",
        contype="0",
        conaffinity="0",
    )
    goal.joint(
        name="goal_slidey",
        type="slide",
        pos="0 0 0",
        axis="0 1 0",
        range="-10.3213 10.3",
        damping="0.5",
    )
    goal.joint(
        name="goal_slidex",
        type="slide",
        pos="0 0 0",
        axis="1 0 0",
        range="-10.3213 10.3",
        damping="0.5",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(
        joint="r_shoulder_pan_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True
    )
    actuator.motor(
        joint="r_shoulder_lift_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True
    )
    actuator.motor(
        joint="r_upper_arm_roll_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True
    )
    actuator.motor(joint="r_elbow_flex_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True)
    actuator.motor(
        joint="r_forearm_roll_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True
    )
    actuator.motor(joint="r_wrist_flex_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True)
    actuator.motor(joint="r_wrist_roll_joint", ctrlrange=[-2.0, 2.0], ctrllimited=True)

    return mjcmodel


def swimmer():
    mjcmodel = MJCModel("swimmer")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="degree", coordinate="local")
    mjcmodel.root.option(
        timestep=0.01,
        viscosity=0.1,
        density=4000,
        integrator="RK4",
        collision="predefined",
    )
    default = mjcmodel.root.default()
    default.joint(armature=0.1)
    default.geom(
        rgba=[0.8, 0.6, 0.1, 1], condim=1, contype=1, conaffinity=1, material="geom"
    )

    asset = mjcmodel.root.asset()
    asset.texture(
        builtin="gradient",
        height=100,
        rgb1=[1, 1, 1],
        rgb2=[0, 0, 0],
        type="skybox",
        width=100,
    )
    asset.texture(
        builtin="flat",
        height=1278,
        mark="cross",
        markrgb=[1, 1, 1],
        name="texgeom",
        random=0.01,
        rgb1=[0.8, 0.6, 0.4],
        rgb2=[0.8, 0.6, 0.4],
        type="cube",
        width=127,
    )
    asset.texture(
        builtin="checker",
        height=100,
        name="texplane",
        rgb1=[0, 0, 0],
        rgb2=[0.8, 0.8, 0.8],
        type="2d",
        width=100,
    )
    asset.material(
        name="MatPlane",
        reflectance=0.5,
        shininess=1,
        specular=1,
        texrepeat=[30, 30],
        texture="texplane",
    )
    asset.material(name="geom", texture="texgeom", texuniform=True)

    worldbody = mjcmodel.root.worldbody()
    worldbody.light(
        cutoff=100,
        diffuse=[1, 1, 1],
        dir=[0, 0, -1.3],
        directional=True,
        exponent=1,
        pos=[0, 0, 1.3],
        specular=[0.1, 0.1, 0.1],
    )
    worldbody.geom(
        conaffinity=1,
        condim=3,
        material="MatPlane",
        name="floor",
        pos=[0, 0, -0.1],
        rgba=[0.8, 0.9, 0.9, 1],
        size=[40, 40, 0.1],
        type="plane",
    )
    torso = worldbody.body(name="torso", pos=[0, 0, 0])
    torso.geom(density=1000, fromto=[1.5, 0, 0, 0.5, 0, 0], size=0.1, type="capsule")
    torso.joint(axis=[1, 0, 0], name="slider1", pos=[0, 0, 0], type="slide")
    torso.joint(axis=[0, 1, 0], name="slider2", pos=[0, 0, 0], type="slide")
    torso.joint(axis=[0, 0, 1], name="rot", pos=[0, 0, 0], type="hinge")
    mid = torso.body(name="mid", pos=[0.5, 0, 0])
    mid.geom(density=1000, fromto=[0, 0, 0, -1, 0, 0], size=0.1, type="capsule")
    mid.joint(
        axis=[0, 0, 1],
        limited=True,
        name="rot2",
        pos=[0, 0, 0],
        range=[-100, 100],
        type="hinge",
    )
    back = mid.body(name="back", pos=[-1, 0, 0])
    back.geom(density=1000, fromto=[0, 0, 0, -1, 0, 0], size=0.1, type="capsule")
    back.joint(
        axis=[0, 0, 1],
        limited=True,
        name="rot3",
        pos=[0, 0, 0],
        range=[-100, 100],
        type="hinge",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(ctrllimited=True, ctrlrange=[-1, 1], gear=150, joint="rot2")
    actuator.motor(ctrllimited=True, ctrlrange=[-1, 1], gear=150, joint="rot3")
    return mjcmodel


def swimmer_rllab():
    mjcmodel = MJCModel("swimmer")
    mjcmodel.root.compiler(inertiafromgeom="true", angle="degree", coordinate="local")
    mjcmodel.root.option(
        timestep=0.01,
        viscosity=0.1,
        density=4000,
        integrator="Euler",
        iterations=1000,
        collision="predefined",
    )

    custom = mjcmodel.root.custom()
    custom.numeric(name="frame_skip", data=50)

    default = mjcmodel.root.default()
    # default.joint(armature=0.1)
    default.geom(
        rgba=[0.8, 0.6, 0.1, 1], condim=1, contype=1, conaffinity=1, material="geom"
    )

    asset = mjcmodel.root.asset()
    asset.texture(
        builtin="gradient",
        height=100,
        rgb1=[1, 1, 1],
        rgb2=[0, 0, 0],
        type="skybox",
        width=100,
    )
    asset.texture(
        builtin="flat",
        height=1278,
        mark="cross",
        markrgb=[1, 1, 1],
        name="texgeom",
        random=0.01,
        rgb1=[0.8, 0.6, 0.4],
        rgb2=[0.8, 0.6, 0.4],
        type="cube",
        width=127,
    )
    asset.texture(
        builtin="checker",
        height=100,
        name="texplane",
        rgb1=[0, 0, 0],
        rgb2=[0.8, 0.8, 0.8],
        type="2d",
        width=100,
    )
    asset.material(
        name="MatPlane",
        reflectance=0.5,
        shininess=1,
        specular=1,
        texrepeat=[30, 30],
        texture="texplane",
    )
    asset.material(name="geom", texture="texgeom", texuniform=True)

    worldbody = mjcmodel.root.worldbody()
    worldbody.light(
        cutoff=100,
        diffuse=[1, 1, 1],
        dir=[0, 0, -1.3],
        directional=True,
        exponent=1,
        pos=[0, 0, 1.3],
        specular=[0.1, 0.1, 0.1],
    )
    worldbody.geom(
        conaffinity=1,
        condim=3,
        material="MatPlane",
        name="floor",
        pos=[0, 0, -0.1],
        rgba=[0.8, 0.9, 0.9, 1],
        size=[40, 40, 0.1],
        type="plane",
    )
    torso = worldbody.body(name="torso", pos=[0, 0, 0])
    torso.geom(density=1000, fromto=[1.5, 0, 0, 0.5, 0, 0], size=0.1, type="capsule")
    torso.joint(axis=[1, 0, 0], name="slider1", pos=[0, 0, 0], type="slide")
    torso.joint(axis=[0, 1, 0], name="slider2", pos=[0, 0, 0], type="slide")
    torso.joint(axis=[0, 0, 1], name="rot", pos=[0, 0, 0], type="hinge")
    mid = torso.body(name="mid", pos=[0.5, 0, 0])
    mid.geom(density=1000, fromto=[0, 0, 0, -1, 0, 0], size=0.1, type="capsule")
    mid.joint(
        axis=[0, 0, 1],
        limited=True,
        name="rot2",
        pos=[0, 0, 0],
        range=[-100, 100],
        type="hinge",
    )
    back = mid.body(name="back", pos=[-1, 0, 0])
    back.geom(density=1000, fromto=[0, 0, 0, -1, 0, 0], size=0.1, type="capsule")
    back.joint(
        axis=[0, 0, 1],
        limited=True,
        name="rot3",
        pos=[0, 0, 0],
        range=[-100, 100],
        type="hinge",
    )

    actuator = mjcmodel.root.actuator()
    actuator.motor(ctrllimited=True, ctrlrange=[-50, 50], joint="rot2")
    actuator.motor(ctrllimited=True, ctrlrange=[-50, 50], joint="rot3")
    return mjcmodel



"""Helper methods to build and run neural networks."""

import collections
import contextlib
from typing import Callable, Dict, Iterable, Optional, Sequence, Tuple

import gym
import tensorflow as tf
from stable_baselines.common.input import observation_input

LayersDict = Dict[str, tf.layers.Layer]


def build_mlp(
    hid_sizes: Iterable[int],
    name: Optional[str] = None,
    activation: Optional[Callable] = tf.nn.relu,
    initializer: Optional[Callable] = None,
) -> LayersDict:
    """Constructs an MLP, returning an ordered dict of layers."""
    layers = collections.OrderedDict()

    # Hidden layers
    for i, size in enumerate(hid_sizes):
        key = f"{name}_dense{i}"
        layer = tf.layers.Dense(
            size, activation=activation, kernel_initializer=initializer, name=key
        )  # type: tf.layers.Layer
        layers[key] = layer

    # Final layer
    layer = tf.layers.Dense(
        1, kernel_initializer=initializer, name=f"{name}_dense_final"
    )  # type: tf.layers.Layer
    layers[f"{name}_dense_final"] = layer

    return layers


def sequential(inputs: tf.Tensor, layers: LayersDict) -> tf.Tensor:
    """Applies a sequence of layers to an input."""
    output = inputs
    for layer in layers.values():
        output = layer(output)
    output = tf.squeeze(output, axis=1)
    return output


def build_inputs(
    observation_space: gym.Space, action_space: gym.Space, scale: bool = False
) -> Tuple[tf.Tensor, ...]:
    """Builds placeholders and processed input Tensors.

    Observation `obs_*` and `next_obs_*` placeholders and processed input
    tensors have shape `(None,) + obs_space.shape`.
    The action `act_*` placeholder and processed input tensors have shape
    `(None,) + act_space.shape`.

    Args:
        observation_space: The observation space.
        action_space: The action space.
        scale: Only relevant for environments with Box spaces. If True, then
            processed input Tensors are automatically scaled to the interval [0, 1].

    Returns:
        (phs, inps) where phs is a tuple of:
            obs_ph: Placeholder for old observations.
            act_ph: Placeholder for actions.
            next_obs_ph: Placeholder for new observations.
            done_ph: Placeholder for boolean episode termination.
        and inps is a tuple of:
            obs_inp: Network-ready float32 Tensor with processed old observations.
            act_inp: Network-ready float32 Tensor with processed actions.
            next_obs_inp: Network-ready float32 Tensor with processed new observations.
            dones_inp: Network-ready float32 tensor, with booleans 0-1 coded.
    """
    obs_ph, obs_inp = observation_input(observation_space, name="obs", scale=scale)
    act_ph, act_inp = observation_input(action_space, name="act", scale=scale)
    next_obs_ph, next_obs_inp = observation_input(
        observation_space, name="next_obs", scale=scale
    )
    done_ph = tf.placeholder(name="dones", shape=(None,), dtype=tf.bool)
    done_inp = tf.cast(done_ph, dtype=tf.float32)
    phs = (obs_ph, act_ph, next_obs_ph, done_ph)
    inps = (obs_inp, act_inp, next_obs_inp, done_inp)
    return phs, inps


@contextlib.contextmanager
def make_session(close_on_exit: bool = True, **kwargs):
    """Context manager for a TensorFlow session.

    The session is associated with a newly created graph. Both session and
    graph are set as default. The session will be closed when exiting this
    context manager.

    Args:
      close_on_exit: If True, closes the session upon leaving the context manager.
      kwargs: passed through to `tf.Session`.

    Yields:
      (graph, session) where graph is a `tf.Graph` and `session` a `tf.Session`.
    """
    graph = tf.Graph()
    with graph.as_default():
        session = tf.Session(graph=graph, **kwargs)
        try:
            with session.as_default():
                yield graph, session
        finally:
            if close_on_exit:
                session.close()


def build_and_apply_mlp(
    inputs: Sequence[tf.Tensor],
    hid_sizes: Optional[Iterable[int]] = None,
    **kwargs: dict,
) -> Tuple[tf.Tensor, LayersDict]:
    """Builds an MLP depending on specified inputs.

    All specified inputs will be flattened and then concatenated. They are then
    applied using `sequential` to an MLP built using `build_mlp`.

    Arguments:
        hid_sizes: Number of units at each hidden layer.
            Default is (32, 32), i.e. two hidden layers with 32 units.
            () represents linear.
      inputs: Sequence of tensor inputs to flatten and concatenate.
      **kwargs: Passed through to `util.build_mlp`.

    Returns:
      (output, mlp) where output is the predicted reward and mlp is a LayersDict.

    Raises:
      ValueError: If inputs is a length-0 tensor.
    """
    if len(inputs) == 0:
        raise ValueError("Must specify at least one input")

    if hid_sizes is None:
        hid_sizes = (32, 32)

    with tf.variable_scope("theta"):
        inputs = [tf.layers.flatten(x) for x in inputs]
        inputs = tf.concat(inputs, axis=1)
        mlp = build_mlp(hid_sizes=hid_sizes, name="reward", **kwargs)
        output = sequential(inputs, mlp)

        return output, mlp



from typing import Optional

from gym.envs import register as gym_register

_ENTRY_POINT_PREFIX = "imitation.envs.examples.airl_envs"


def _register(env_name: str, entry_point: str, kwargs: Optional[dict] = None):
    entry_point = f"{_ENTRY_POINT_PREFIX}.{entry_point}"
    gym_register(id=env_name, entry_point=entry_point, kwargs=kwargs)


def _point_maze_register():
    for dname, dval in {"Left": 0, "Right": 1}.items():
        for vname, vval in {"": False, "Vel": True}.items():
            _register(
                f"imitation/PointMaze{dname}{vname}-v0",
                entry_point="point_maze_env:PointMazeEnv",
                kwargs={"direction": dval, "include_vel": vval},
            )


_register(
    "imitation/ObjPusher-v0",
    entry_point="pusher_env:PusherEnv",
    kwargs={"sparse_reward": False},
)
_register("imitation/TwoDMaze-v0", entry_point="twod_maze:TwoDMaze")

_point_maze_register()

# A modified ant which flips over less and learns faster via TRPO
_register(
    "imitation/CustomAnt-v0",
    entry_point="ant_env:CustomAntEnv",
    kwargs={"gear": 30, "disabled": False},
)
_register(
    "imitation/DisabledAnt-v0",
    entry_point="ant_env:CustomAntEnv",
    kwargs={"gear": 30, "disabled": True},
)



# pytype: skip-file
# flake8: noqa
import abc
import warnings
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

import gym
import numpy as np

from imitation.contrib import dataset_wrappers
from imitation.data import buffer, datasets, types

S = TypeVar("S")
T = TypeVar("T")


class DatasetWrapper(Generic[S, T], datasets.Dataset[T]):
    """Abstract base class for classes that wrap other Dataset instances.

    Concrete subclasses should implement `.sample(n_samples) -> T`.
    """

    def __init__(self, dataset: datasets.Dataset[S], name: Optional[str] = None):
        self.dataset = dataset
        self.name = name

    def size(self):
        return self.dataset.size()

    def __str__(self):
        name = self.name or type(self).__name__
        return f"<{name}{self.dataset}>"

    def __repr__(self):
        return str(self)


class ReplayBufferDatasetWrapper(DatasetWrapper):
    def __init__(
        self,
        dataset: datasets.Dataset[Dict[str, np.ndarray]],
        capacity: int,
    ):
        """Wrapper that modifies .sample() to sample indirectly using a replay buffer.

        With every call to `.sample(n_samples)`, `n_samples` samples from the wrapped
        Dataset are stored in a FIFO buffer, and `n_samples` samples are randomly
        selected and returned from the buffer.

        Args:
            dataset: A Dataset returning `TransitionsWithRew` samples.
            capacity: The number of samples that can be stored in the `ReplayBuffer`.
                Upon initialization of this wrapper, the `ReplayBuffer` is immediately
                populated with `capacity` samples from `dataset.sample(capacity).
        """

        super().__init__(dataset, f"ReplayBuffer[cap={capacity}]")
        self.capacity = capacity
        samples = self.dataset.sample(self.capacity)
        self.buffer = buffer.ReplayBuffer.from_data(samples)

    def sample(self, n_samples: int):
        assert n_samples >= 0
        self.buffer.store(n_samples, truncate_ok=True)
        return self.buffer.sample(n_samples)


class BidirectionalTransform(abc.ABC):
    def __init__(self, observation_space: gym.Space, action_space: gym.Space):
        self.observation_space = observation_space
        self.action_space = action_space  # Processed action space.

    @abc.abstractmethod
    def forwards(
        self,
        transitions: types.TransitionsWithRew,
    ) -> types.TransitionsWithRew:
        """Defines a transformation from raw (Dataset or Env) to transformed values.

        `forwards` should be a pure function. In other words, any instance of
        BidirectionalTransform returns the same outputs given the same inputs.

        Args:
            ob: Raw observation (as passed to `Env.step(act)` or `Dataset.sample()`).
            act: Raw action (as returned from `Dataset.sample()` or passed to
                `Env.step(act)`.
            next_ob: Raw observation for the next timestep. Can be None, in which
                case the corresponding return value should also be None.
            rew: Raw reward (as returned from `Env.step()` or from `Dataset.sample()`.
            info: Raw info (as returned from `Env.step()` or from `Dataset.sample()`.

        Returns:
            Transformed versions of the arguments.
        """

    @abc.abstractmethod
    def backwards_act(self, transformed_act):
        """Defines a (pseudo-)inverse transformation from transformed act to raw act.

        `backwards_act` should be a pure function. In other words, any instance of
        BidirectionalTransform returns the same outputs given the same inputs.


        Args:
            transformed_act: Actions transformed from raw actions by `forwards()`.

        Returns:
            A raw action. Not required to be the same raw action as that passed into
            `forwards()` because `forwards()` might not be one-to-one / injective.
        """


class DefinesDatasetWrapperMixin(abc.ABC, Generic[S, T]):
    """A class implementing this mixin defines an `DatasetWrapper`."""

    @abc.abstractmethod
    def apply_dataset_wrapper(
        self, dataset: datasets.Dataset[S]
    ) -> DatasetWrapper[S, T]:
        """Wraps a dataset using the dataset.

        Args:
            dataset: A Dataset to wrap.

        Returns:
            A wrapped dataset.
        """


class BidirectionalGymWrapper(
    gym.Wrapper,
    DefinesDatasetWrapperMixin[types.TransitionsWithRew, types.TransitionsWithRew],
):
    def __init__(self, env: gym.Env, transform: BidirectionalTransform):
        self.transform = transform
        super().__init__(env)

    @property
    def observation_space(self) -> gym.Space:
        return self.transform.observation_space

    @property
    def action_space(self) -> gym.Space:
        return self.transform.action_space

    def step(self, action: Any) -> Tuple[Any, float, dict]:
        act = self.transform.backwards_act(action)
        return self.env.step(act)

    def apply_dataset_wrapper(
        self, dataset: datasets.Dataset[types.TransitionsWithRew]
    ) -> DatasetWrapper[types.TransitionsWithRew, types.TransitionsWithRew]:
        """Wraps a dataset using the dataset."""
        return BidirectionalDatasetWrapper(dataset, self.transform)


class BidirectionalDatasetWrapper(
    DatasetWrapper[types.TransitionsWithRew, types.TransitionsWithRew],
):
    def __init__(
        self,
        dataset: datasets.Dataset[types.TransitionsWithRew],
        transform: BidirectionalTransform,
    ):
        self.dataset = dataset
        self.transform = transform
        super().__init__(dataset)

    def sample(self, n_samples: int) -> types.TransitionsWithRew:
        samples = self.dataset.sample(n_samples)
        iterator = zip(
            samples.obs,
            samples.acts,
            samples.rews,
            samples.dones,
            samples.infos,
            samples.next_obs,
        )
        transformed_iterator = map(self.transform.forwards, iterator)
        transformed_data_needs_stack = list(zip(*transformed_iterator))
        transformed_data = tuple(map(np.stack, transformed_data_needs_stack))
        obs, acts, rews, dones, infos, next_obs = transformed_data
        transformed_samples = types.TransitionsWithRew(
            obs=obs,
            acts=acts,
            infos=infos,
            next_obs=next_obs,
            dones=dones,
            rews=rews,
        )
        return transformed_samples


def wrap_dataset_with_env_wrappers(
    dataset: datasets.Dataset[types.TransitionsWithRew],
    env: gym.Env,
    warn_on_ignore_wrapper: bool = True,
) -> datasets.Dataset[types.TransitionsWithRew]:
    """Apply GymDatasetWrappers on top of a Dataset[TransitionsWithRew].

    Args:
        dataset: ...
        env: An Env possibly with wrappers. For every gym.Wrapper wrapping `env` which
            also subclasses `GymWrapperDatasetMixin`, `minerl_dataset` is wrapped with
            MineRLDatasetWrapper which applies that the gym.Wrapper's transform
            to the dataset, as defined in `GymWrapperDatasetMixin.process_batch()`.

            Any `gym.Wrapper` that doesn't subclass `GymWrapperDatasetMixin` is skipped.
        warn_on_ignore_wrapper: If True, then warn with RuntimeWarning for every
            `gym.Wrapper` that doesn't subclass `GymWrapperDatasetMixin`.

    Returns:
        `dataset` wrapped by a `DatasetWrapper` for every gym.Wrapper around `env`
        also subclasses `GymWrapperDatasetMixin`. If `env` has no wrappers, then the
        return is simply `dataset`.
    """
    # Traverse all the gym.Wrapper starting from the outermost wrapper. When re-apply
    # them to the Dataset as MineRLDatasetWrapper, we need to apply in reverse order
    # (starting with innermost wrapper), so we save gym.Wrappers in a list first.
    compatible_wrappers: List[DefinesDatasetWrapperMixin] = []
    curr = env
    while isinstance(curr, gym.Wrapper):
        if isinstance(curr, DefinesDatasetWrapperMixin):
            compatible_wrappers.append(curr)
        elif warn_on_ignore_wrapper:
            warnings.warn(
                f"{curr} doesn't subclass DefinesDatasetWrapperMixin. Skipping "
                "this Wrapper when creating Dataset processing wrappers.",
                RuntimeWarning,
            )
        curr = curr.env

    for wrapper in reversed(dataset_wrappers):
        dataset = DatasetWrapper(dataset, wrapper)
    return dataset



"""Load serialized reward functions of different types."""

import contextlib
from typing import Callable, ContextManager, Iterator

import numpy as np
from stable_baselines.common.vec_env import VecEnv

from imitation.rewards import common, discrim_net, reward_net
from imitation.util import networks, registry, util

RewardFnLoaderFn = Callable[[str, VecEnv], ContextManager[common.RewardFn]]

reward_registry: registry.Registry[RewardFnLoaderFn] = registry.Registry()


@registry.sess_context
def _load_discrim_net(path: str, venv: VecEnv) -> common.RewardFn:
    """Load test reward output from discriminator."""
    del venv  # Unused.
    discriminator = discrim_net.DiscrimNet.load(path)
    # TODO(gleave): expose train reward as well? (hard due to action probs?)
    return discriminator.reward_test


def _load_reward_net_as_fn(shaped: bool) -> RewardFnLoaderFn:
    @contextlib.contextmanager
    def loader(path: str, venv: VecEnv) -> Iterator[common.RewardFn]:
        """Load train (shaped) or test (not shaped) reward from path."""
        del venv  # Unused.
        with networks.make_session() as (graph, sess):
            net = reward_net.RewardNet.load(path)
            reward = net.reward_output_train if shaped else net.reward_output_test

            def rew_fn(
                obs: np.ndarray,
                act: np.ndarray,
                next_obs: np.ndarray,
                dones: np.ndarray,
            ) -> np.ndarray:
                fd = {
                    net.obs_ph: obs,
                    net.act_ph: act,
                    net.next_obs_ph: next_obs,
                    net.done_ph: dones,
                }
                rew = sess.run(reward, feed_dict=fd)
                assert rew.shape == (len(obs),)
                return rew

            yield rew_fn

    return loader


def load_zero(path: str, venv: VecEnv) -> common.RewardFn:
    del path, venv

    def f(
        obs: np.ndarray, act: np.ndarray, next_obs: np.ndarray, dones: np.ndarray
    ) -> np.ndarray:
        del act, next_obs, dones  # Unused.
        return np.zeros(obs.shape[0])

    return f


reward_registry.register(key="DiscrimNet", value=_load_discrim_net)
reward_registry.register(
    key="RewardNet_shaped", value=_load_reward_net_as_fn(shaped=True)
)
reward_registry.register(
    key="RewardNet_unshaped", value=_load_reward_net_as_fn(shaped=False)
)
reward_registry.register(key="zero", value=registry.dummy_context(load_zero))


@util.docstring_parameter(reward_types=", ".join(reward_registry.keys()))
def load_reward(
    reward_type: str, reward_path: str, venv: VecEnv
) -> ContextManager[common.RewardFn]:
    """Load serialized policy.

    Args:
      reward_type: A key in `reward_registry`. Valid types
          include {reward_types}.
      reward_path: A path specifying the reward.
      venv: An environment that the policy is to be used with.
    """
    reward_loader = reward_registry.get(reward_type)
    return reward_loader(reward_path, venv)



import os

import sacred

from imitation.util import util

eval_policy_ex = sacred.Experiment("eval_policy")


@eval_policy_ex.config
def replay_defaults():
    env_name = "CartPole-v1"  # environment to evaluate in
    eval_n_timesteps = int(1e4)  # Min timesteps to evaluate, optional.
    eval_n_episodes = None  # Num episodes to evaluate, optional.
    num_vec = 1  # number of environments in parallel
    parallel = False  # Use SubprocVecEnv (generally faster if num_vec>1)
    max_episode_steps = None  # Set to positive int to limit episode horizons

    render = True  # render to screen
    render_fps = 60  # -1 to render at full speed
    log_root = os.path.join("output", "eval_policy")  # output directory

    policy_type = "ppo2"  # class to load policy, see imitation.policies.loader
    policy_path = (
        "tests/data/expert_models/" "cartpole_0/policies/final/"
    )  # serialized policy

    reward_type = None  # Optional: override with reward of this type
    reward_path = None  # Path of serialized reward to load


@eval_policy_ex.config
def logging(log_root, env_name):
    log_dir = os.path.join(
        log_root, env_name.replace("/", "_"), util.make_unique_timestamp()
    )


@eval_policy_ex.named_config
def fast():
    eval_n_timesteps = int(1e2)
    eval_n_episodes = None
    max_episode_steps = int(1e2)



import numpy as np
from gym import utils
from gym.envs.mujoco import mujoco_env

from imitation.envs.examples.airl_envs.dynamic_mjc.mjc_models import pusher


class PusherEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self, sparse_reward=False, no_reward=False, episode_length=200):
        utils.EzPickle.__init__(self)
        self.sparse_reward = sparse_reward
        self.no_reward = no_reward
        self.max_episode_length = episode_length
        self.goal_pos = np.asarray([0.0, 0.0])

        self.episode_length = 0

        model = pusher(goal_pos=[self.goal_pos[0], self.goal_pos[1], -0.323])
        with model.asfile() as f:
            mujoco_env.MujocoEnv.__init__(self, f.name, 5)

    def step(self, a):
        vec_1 = self.get_body_com("object") - self.get_body_com("tips_arm")
        vec_2 = self.get_body_com("object") - self.get_body_com("goal")

        reward_near = -np.linalg.norm(vec_1)  # arm to object
        reward_dist = -np.linalg.norm(vec_2[0:2])  # object to goal
        reward_ctrl = -np.square(a).sum()
        if self.no_reward:
            reward = 0
        elif self.sparse_reward:
            reward = reward_dist + 0.1 * reward_ctrl
        else:
            reward = reward_dist + 0.1 * reward_ctrl + 0.5 * reward_near

        self.do_simulation(a, self.frame_skip)
        ob = self._get_obs()
        self.episode_length += 1
        done = self.episode_length >= self.max_episode_length
        return (
            ob,
            reward,
            done,
            dict(
                reward_dist=reward_dist,
                reward_near=reward_near,
                reward_ctrl=reward_ctrl,
            ),
        )

    def viewer_setup(self):
        self.viewer.cam.trackbodyid = -1
        self.viewer.cam.distance = 4.0

    def reset_model(self):
        qpos = self.init_qpos
        self.episode_length = 0

        while True:
            self.cylinder_pos = np.concatenate(
                [
                    self.np_random.uniform(low=-0.5, high=0, size=1),
                    self.np_random.uniform(low=-0.5, high=0.5, size=1),
                ]
            )
            cyl_dist = np.linalg.norm(self.cylinder_pos - self.goal_pos)
            if cyl_dist > 0.2 and cyl_dist < 0.4:
                break

        qpos[-4:-2] = self.cylinder_pos
        qpos[-2] = self.goal_pos[0]
        qpos[-1] = self.goal_pos[1]
        qvel = self.init_qvel + self.np_random.uniform(
            low=-0.005, high=0.005, size=self.model.nv
        )
        qvel[-4:] = 0
        self.set_state(qpos, qvel)
        return self._get_obs()

    def _get_obs(self):
        return np.concatenate(
            [
                self.sim.data.qpos.flat[:7],
                self.sim.data.qvel.flat[:7],
                self.get_body_com("tips_arm"),
                self.get_body_com("object"),
                self.get_body_com("goal"),
            ]
        )

    def plot_trajs(self, *args, **kwargs):
        pass



"""Common wrapper for adding custom reward values to an environment."""
import collections

import numpy as np
from stable_baselines.common import vec_env

from imitation.rewards import common


class RewardVecEnvWrapper(vec_env.VecEnvWrapper):
    def __init__(
        self, venv: vec_env.VecEnv, reward_fn: common.RewardFn, ep_history: int = 100
    ):
        """Uses a provided reward_fn to replace the reward function returned by `step()`.

        Automatically resets the inner VecEnv upon initialization. A tricky part
        about this class is keeping track of the most recent observation from each
        environment.

        Will also include the previous reward given by the inner VecEnv in the
        returned info dict under the `wrapped_env_rew` key.

        Args:
            venv: The VecEnv to wrap.
            reward_fn: A function that wraps takes in vectorized transitions
                (obs, act, next_obs) a vector of episode timesteps, and returns a
                vector of rewards.
            ep_history: The number of episode rewards to retain for computing
                mean reward.
        """
        assert not isinstance(venv, RewardVecEnvWrapper)
        super().__init__(venv)
        self.episode_rewards = collections.deque(maxlen=ep_history)
        self._cumulative_rew = np.zeros((venv.num_envs,))
        self.reward_fn = reward_fn
        self.reset()

    def log_callback(self, logger):
        """Logs mean reward over the last `ep_history` episodes."""
        if len(self.episode_rewards) == 0:
            return
        mean = sum(self.episode_rewards) / len(self.episode_rewards)
        logger.logkv("eprewmean_wrapped", mean)

    @property
    def envs(self):
        return self.venv.envs

    def reset(self):
        self._old_obs = self.venv.reset()
        return self._old_obs

    def step_async(self, actions):
        self._actions = actions
        return self.venv.step_async(actions)

    def step_wait(self):
        obs, old_rews, dones, infos = self.venv.step_wait()

        # The vecenvs automatically reset the underlying environments once they
        # encounter a `done`, in which case the last observation corresponding to
        # the `done` is dropped. We're going to pull it back out of the info dict!
        obs_fixed = []
        for single_obs, single_done, single_infos in zip(obs, dones, infos):
            if single_done:
                single_obs = single_infos["terminal_observation"]

            obs_fixed.append(single_obs)
        obs_fixed = np.stack(obs_fixed)

        rews = self.reward_fn(self._old_obs, self._actions, obs_fixed, np.array(dones))
        assert len(rews) == len(obs), "must return one rew for each env"
        done_mask = np.asarray(dones, dtype="bool").reshape((len(dones),))

        # Update statistics
        self._cumulative_rew += rews
        for single_done, single_ep_rew in zip(dones, self._cumulative_rew):
            if single_done:
                self.episode_rewards.append(single_ep_rew)
        self._cumulative_rew[done_mask] = 0

        # we can just use obs instead of obs_fixed because on the next iteration
        # after a reset we DO want to access the first observation of the new
        # trajectory, not the last observation of the old trajectory
        self._old_obs = obs
        for info_dict, old_rew in zip(infos, old_rews):
            info_dict["wrapped_env_rew"] = old_rew
        return obs, rews, dones, infos



"""Finite-horizon discrete environments with known transition dynamics.

These are handy when you want to perform exact maxent policy optimisation.
"""

import abc

import gym
import numpy as np
from gym import spaces


class ResettableEnv(gym.Env, abc.ABC):
    """ABC for environments that are resettable.

    Specifically, these environments provide oracle access to sample from the initial
    state distribution and transition dynamics, and compute the reward and termination
    condition. Almost all simulated environments can meet these criteria.
    """

    def __init__(self):
        self._state_space = None
        self._observation_space = None
        self._action_space = None
        self.cur_state = None
        self._n_actions_taken = None
        self.seed()

    @abc.abstractmethod
    def initial_state(self):
        """Samples from the initial state distribution."""

    @abc.abstractmethod
    def transition(self, state, action):
        """Samples from transition distribution."""

    @abc.abstractmethod
    def reward(self, state, action, new_state):
        """Computes reward for a given transition."""

    @abc.abstractmethod
    def terminal(self, state, step: int) -> bool:
        """Is the state terminal?"""

    @abc.abstractmethod
    def obs_from_state(self, state):
        """Returns observation produced by a given state."""

    @property
    def state_space(self) -> gym.Space:
        """State space.

        Often same as observation_space, but differs in POMDPs.
        """
        return self._state_space

    @property
    def observation_space(self) -> gym.Space:
        """Observation space.

        Return type of reset() and component of step().
        """
        return self._observation_space

    @property
    def action_space(self) -> gym.Space:
        """Action space.

        Parameter type of step().
        """
        return self._action_space

    @property
    def n_actions_taken(self) -> int:
        """Number of steps taken so far."""
        return self._n_actions_taken

    def seed(self, seed=None):
        if seed is None:
            # Gym API wants list of seeds to be returned for some reason, so
            # generate a seed explicitly in this case
            seed = np.random.randint(0, 1 << 31)
        self.rand_state = np.random.RandomState(seed)
        return [seed]

    def reset(self):
        self.cur_state = self.initial_state()
        self._n_actions_taken = 0
        return self.obs_from_state(self.cur_state)

    def step(self, action):
        if self.cur_state is None or self._n_actions_taken is None:
            raise ValueError("Need to call reset() before first step()")

        old_state = self.cur_state
        self.cur_state = self.transition(self.cur_state, action)
        obs = self.obs_from_state(self.cur_state)
        rew = self.reward(old_state, action, self.cur_state)
        done = self.terminal(self.cur_state, self._n_actions_taken)
        self._n_actions_taken += 1

        infos = {"old_state": old_state, "new_state": self.cur_state}
        return obs, rew, done, infos


class TabularModelEnv(ResettableEnv, abc.ABC):
    """ABC for tabular environments with known dynamics."""

    def __init__(self):
        """Initialise common attributes of all model-based environments.

        Attributes include current state & number of actions taken so far (initial
        None, so that error can be thrown if reset() is not called), attributes for
        cached observation/action space, and random seed for rollouts."""
        super().__init__()

    @property
    def state_space(self) -> gym.Space:
        # Construct spaces lazily, so they can depend on properties in subclasses.
        if self._state_space is None:
            self._state_space = spaces.Discrete(self.state_dim)
        return self._state_space

    @property
    def observation_space(self) -> gym.Space:
        # Construct spaces lazily, so they can depend on properties in subclasses.
        if self._observation_space is None:
            self._observation_space = spaces.Box(
                low=float("-inf"), high=float("inf"), shape=(self.obs_dim,)
            )
        return self._observation_space

    @property
    def action_space(self) -> gym.Space:
        # Construct spaces lazily, so they can depend on properties in subclasses.
        if self._action_space is None:
            self._action_space = spaces.Discrete(self.n_actions)
        return self._action_space

    def initial_state(self):
        return self.rand_state.choice(self.n_states, p=self.initial_state_dist)

    def transition(self, state, action):
        out_dist = self.transition_matrix[state, action]
        choice_states = np.arange(self.n_states)
        return int(self.rand_state.choice(choice_states, p=out_dist, size=()))

    def reward(self, state, action, new_state):
        reward = self.reward_matrix[state]
        assert np.isscalar(reward), reward
        return reward

    def terminal(self, state, n_actions_taken):
        return n_actions_taken >= self.horizon

    def obs_from_state(self, state):
        # Copy so it can't be mutated in-place (updates will be reflected in
        # self.observation_matrix!)
        obs = self.observation_matrix[state].copy()
        assert obs.ndim == 1, obs.shape
        return obs

    @property
    def n_states(self):
        """Number of states in this MDP (int)."""
        return self.transition_matrix.shape[0]

    @property
    def n_actions(self):
        """Number of actions in this MDP (int)."""
        return self.transition_matrix.shape[1]

    @property
    def state_dim(self):
        """Size of state vectors for this MDP."""
        return self.observation_matrix.shape[0]

    @property
    def obs_dim(self):
        """Size of observation vectors for this MDP."""
        return self.observation_matrix.shape[1]

    # ############################### #
    # METHODS THAT MUST BE OVERRIDDEN #
    # ############################### #

    @property
    @abc.abstractmethod
    def transition_matrix(self):
        """3D transition matrix.

        Dimensions correspond to current state, current action, and next state.

        In other words, if `T` is our returned matrix, then `T[s,a,sprime]` is the
        chance of transitioning into state `sprime` after taking action `a` in state
        `s`.
        """

    @property
    @abc.abstractmethod
    def observation_matrix(self):
        """2D observation matrix.

        Dimensions correspond to current state (first dim) and elements of observation
        (second dim)."""

    @property
    @abc.abstractmethod
    def reward_matrix(self):
        """1D reward matrix with an element corresponding to each state."""

    @property
    @abc.abstractmethod
    def horizon(self):
        """Number of actions that can be taken in an episode."""

    @property
    @abc.abstractmethod
    def initial_state_dist(self):
        """1D vector representing a distribution over initial states."""
        return



import collections
import dataclasses
import functools
from typing import Callable, Dict, Hashable, List, Optional, Sequence, Union

import numpy as np
import tensorflow as tf
from stable_baselines.common.base_class import BaseRLModel
from stable_baselines.common.policies import BasePolicy
from stable_baselines.common.vec_env import VecEnv

from imitation.data import types
from imitation.policies.base import get_action_policy


def unwrap_traj(traj: types.TrajectoryWithRew) -> types.TrajectoryWithRew:
    """Uses `RolloutInfoWrapper`-captured `obs` and `rews` to replace fields.

    This can be useful for bypassing other wrappers to retrieve the original
    `obs` and `rews`.

    Fails if `infos` is None or if the trajectory was generated from an
    environment without imitation.util.rollout.RolloutInfoWrapper

    Args:
      traj: A trajectory generated from `RolloutInfoWrapper`-wrapped Environments.

    Returns:
      A copy of `traj` with replaced `obs` and `rews` fields.
    """
    ep_info = traj.infos[-1]["rollout"]
    res = dataclasses.replace(traj, obs=ep_info["obs"], rews=ep_info["rews"])
    assert len(res.obs) == len(res.acts) + 1
    assert len(res.rews) == len(res.acts)
    return res


class TrajectoryAccumulator:
    """Accumulates trajectories step-by-step.

    Useful for collecting completed trajectories while ignoring partially-completed
    trajectories (e.g. when rolling out a VecEnv to collect a set number of
    transitions). Each in-progress trajectory is identified by a 'key', which enables
    several independent trajectories to be collected at once. They key can also be left
    at its default value of `None` if you only wish to collect one trajectory.
    """

    def __init__(self):
        """Initialise the trajectory accumulator."""
        self.partial_trajectories = collections.defaultdict(list)

    def add_step(self, step_dict: Dict[str, np.ndarray], key: Hashable = None):
        """Add a single step to the partial trajectory identified by `key`.

        Generally a single step could correspond to, e.g., one environment managed
        by a VecEnv.

        Args:
            step_dict: dictionary containing information for the current step. Its
                keys could include any (or all) attributes of a `TrajectoryWithRew`
                (e.g. "obs", "acts", etc.).
            key: key to uniquely identify the trajectory to append to, if working
                with multiple partial trajectories.
        """
        self.partial_trajectories[key].append(step_dict)

    def finish_trajectory(self, key: Hashable = None) -> types.TrajectoryWithRew:
        """Complete the trajectory labelled with `key`.

        Args:
            key: key uniquely identifying which in-progress trajectory to remove.

        Returns:
            traj: list of completed trajectories popped from
                `self.partial_trajectories`.
        """
        part_dicts = self.partial_trajectories[key]
        del self.partial_trajectories[key]
        out_dict_unstacked = collections.defaultdict(list)
        for part_dict in part_dicts:
            for key, array in part_dict.items():
                out_dict_unstacked[key].append(array)
        out_dict_stacked = {
            key: np.stack(arr_list, axis=0)
            for key, arr_list in out_dict_unstacked.items()
        }
        traj = types.TrajectoryWithRew(**out_dict_stacked)
        assert traj.rews.shape[0] == traj.acts.shape[0] == traj.obs.shape[0] - 1
        return traj

    def add_steps_and_auto_finish(
        self,
        acts: np.ndarray,
        obs: np.ndarray,
        rews: np.ndarray,
        dones: np.ndarray,
        infos: List[dict],
    ) -> List[types.TrajectoryWithRew]:
        """Calls `add_step` repeatedly using acts and the returns from `venv.step`.

        Also automatically calls `finish_trajectory()` for each `done == True`.
        Before calling this method, each environment index key needs to be
        initialized with the initial observation (usually from `venv.reset()`).

        See the body of `util.rollout.generate_trajectory` for an example.

        Args:
            acts: Actions passed into `VecEnv.step()`.
            obs: Return value from `VecEnv.step(acts)`.
            rews: Return value from `VecEnv.step(acts)`.
            dones: Return value from `VecEnv.step(acts)`.
            infos: Return value from `VecEnv.step(acts)`.
        Returns:
            A list of completed trajectories. There should be one trajectory for
            each `True` in the `dones` argument.
        """
        trajs = []
        for env_idx in range(len(obs)):
            assert env_idx in self.partial_trajectories
            assert list(self.partial_trajectories[env_idx][0].keys()) == ["obs"], (
                "Need to first initialize partial trajectory using "
                "self._traj_accum.add_step({'obs': ob}, key=env_idx)"
            )

        zip_iter = enumerate(zip(acts, obs, rews, dones, infos))
        for env_idx, (act, ob, rew, done, info) in zip_iter:
            if done:
                # actual obs is inaccurate, so we use the one inserted into step info
                # by stable baselines wrapper
                real_ob = info["terminal_observation"]
            else:
                real_ob = ob

            self.add_step(
                dict(
                    acts=act,
                    rews=rew,
                    # this is not the obs corresponding to `act`, but rather the obs
                    # *after* `act` (see above)
                    obs=real_ob,
                    infos=info,
                ),
                env_idx,
            )
            if done:
                # finish env_idx-th trajectory
                new_traj = self.finish_trajectory(env_idx)
                trajs.append(new_traj)
                self.add_step(dict(obs=ob), env_idx)
        return trajs


GenTrajTerminationFn = Callable[[Sequence[types.TrajectoryWithRew]], bool]


def min_episodes(n: int) -> GenTrajTerminationFn:
    """Terminate after collecting n episodes of data.

    Arguments:
      n: Minimum number of episodes of data to collect.
         May overshoot if two episodes complete simultaneously (unlikely).

    Returns:
      A function implementing this termination condition.
    """
    assert n >= 1
    return lambda trajectories: len(trajectories) >= n


def min_timesteps(n: int) -> GenTrajTerminationFn:
    """Terminate at the first episode after collecting n timesteps of data.

    Arguments:
      n: Minimum number of timesteps of data to collect.
        May overshoot to nearest episode boundary.

    Returns:
      A function implementing this termination condition.
    """
    assert n >= 1

    def f(trajectories: Sequence[types.TrajectoryWithRew]):
        timesteps = sum(len(t.obs) - 1 for t in trajectories)
        return timesteps >= n

    return f


def make_sample_until(
    n_timesteps: Optional[int],
    n_episodes: Optional[int],
) -> GenTrajTerminationFn:
    """Returns a termination condition sampling until n_timesteps or n_episodes.

    Arguments:
      n_timesteps: Minimum number of timesteps to sample.
      n_episodes: Number of episodes to sample.

    Returns:
      A termination condition.

    Raises:
      ValueError if both or neither of n_timesteps and n_episodes are set,
      or if either are non-positive.
    """
    if n_timesteps is not None and n_episodes is not None:
        raise ValueError("n_timesteps and n_episodes were both set")
    elif n_timesteps is not None:
        assert n_timesteps > 0
        return min_timesteps(n_timesteps)
    elif n_episodes is not None:
        assert n_episodes > 0
        return min_episodes(n_episodes)
    else:
        raise ValueError("Set at least one of n_timesteps and n_episodes")


def generate_trajectories(
    policy,
    venv: VecEnv,
    sample_until: GenTrajTerminationFn,
    *,
    deterministic_policy: bool = False,
    rng: np.random.RandomState = np.random,
) -> Sequence[types.TrajectoryWithRew]:
    """Generate trajectory dictionaries from a policy and an environment.

    Args:
      policy (BasePolicy or BaseRLModel): A stable_baselines policy or RLModel,
          trained on the gym environment.
      venv: The vectorized environments to interact with.
      sample_until: A function determining the termination condition.
          It takes a sequence of trajectories, and returns a bool.
          Most users will want to use one of `min_episodes` or `min_timesteps`.
      deterministic_policy: If True, asks policy to deterministically return
          action. Note the trajectories might still be non-deterministic if the
          environment has non-determinism!
      rng: used for shuffling trajectories.

    Returns:
      Sequence of trajectories, satisfying `sample_until`. Additional trajectories
      may be collected to avoid biasing process towards short episodes; the user
      should truncate if required.
    """
    if isinstance(policy, BaseRLModel):
        get_action = policy.predict
        policy.set_env(venv)
    else:
        get_action = functools.partial(get_action_policy, policy)

    # Collect rollout tuples.
    trajectories = []
    # accumulator for incomplete trajectories
    trajectories_accum = TrajectoryAccumulator()
    obs = venv.reset()
    for env_idx, ob in enumerate(obs):
        # Seed with first obs only. Inside loop, we'll only add second obs from
        # each (s,a,r,s') tuple, under the same "obs" key again. That way we still
        # get all observations, but they're not duplicated into "next obs" and
        # "previous obs" (this matters for, e.g., Atari, where observations are
        # really big).
        trajectories_accum.add_step(dict(obs=ob), env_idx)

    # Now, we sample until `sample_until(trajectories)` is true.
    # If we just stopped then this would introduce a bias towards shorter episodes,
    # since longer episodes are more likely to still be active, i.e. in the process
    # of being sampled from. To avoid this, we continue sampling until all epsiodes
    # are complete.
    #
    # To start with, all environments are active.
    active = np.ones(venv.num_envs, dtype=np.bool)
    while np.any(active):
        acts, _ = get_action(obs, deterministic=deterministic_policy)
        obs, rews, dones, infos = venv.step(acts)

        # If an environment is inactive, i.e. the episode completed for that
        # environment after `sample_until(trajectories)` was true, then we do
        # *not* want to add any subsequent trajectories from it. We avoid this
        # by just making it never done.
        dones &= active

        new_trajs = trajectories_accum.add_steps_and_auto_finish(
            acts, obs, rews, dones, infos
        )
        trajectories.extend(new_trajs)

        if sample_until(trajectories):
            # Termination condition has been reached. Mark as inactive any environments
            # where a trajectory was completed this timestep.
            active &= ~dones

    # Note that we just drop partial trajectories. This is not ideal for some
    # algos; e.g. BC can probably benefit from partial trajectories, too.

    # Each trajectory is sampled i.i.d.; however, shorter episodes are added to
    # `trajectories` sooner. Shuffle to avoid bias in order. This is important
    # when callees end up truncating the number of trajectories or transitions.
    # It is also cheap, since we're just shuffling pointers.
    rng.shuffle(trajectories)

    # Sanity checks.
    for trajectory in trajectories:
        n_steps = len(trajectory.acts)
        # extra 1 for the end
        exp_obs = (n_steps + 1,) + venv.observation_space.shape
        real_obs = trajectory.obs.shape
        assert real_obs == exp_obs, f"expected shape {exp_obs}, got {real_obs}"
        exp_act = (n_steps,) + venv.action_space.shape
        real_act = trajectory.acts.shape
        assert real_act == exp_act, f"expected shape {exp_act}, got {real_act}"
        exp_rew = (n_steps,)
        real_rew = trajectory.rews.shape
        assert real_rew == exp_rew, f"expected shape {exp_rew}, got {real_rew}"

    return trajectories


def rollout_stats(trajectories: Sequence[types.TrajectoryWithRew]) -> Dict[str, float]:
    """Calculates various stats for a sequence of trajectories.

    Args:
        trajectories: Sequence of trajectories.

    Returns:
        Dictionary containing `n_traj` collected (int), along with episode return
        statistics (keys: `{monitor_,}return_{min,mean,std,max}`, float values)
        and trajectory length statistics (keys: `len_{min,mean,std,max}`, float
        values).

        `return_*` values are calculated from environment rewards.
        `monitor_*` values are calculated from Monitor-captured rewards, and
        are only included if the `trajectories` contain Monitor infos.
    """
    assert len(trajectories) > 0
    out_stats: Dict[str, float] = {"n_traj": len(trajectories)}
    traj_descriptors = {
        "return": np.asarray([sum(t.rews) for t in trajectories]),
        "len": np.asarray([len(t.rews) for t in trajectories]),
    }

    infos_peek = trajectories[0].infos
    if infos_peek is not None and "episode" in infos_peek[-1]:
        monitor_ep_returns = [t.infos[-1]["episode"]["r"] for t in trajectories]
        traj_descriptors["monitor_return"] = np.asarray(monitor_ep_returns)

    stat_names = ["min", "mean", "std", "max"]
    for desc_name, desc_vals in traj_descriptors.items():
        for stat_name in stat_names:
            stat_value: np.generic = getattr(np, stat_name)(desc_vals)
            # Convert numpy type to float or int. The numpy operators always return
            # a numpy type, but we want to return type float. (int satisfies
            # float type for the purposes of static-typing).
            out_stats[f"{desc_name}_{stat_name}"] = stat_value.item()

    for v in out_stats.values():
        assert isinstance(v, (int, float))
    return out_stats


def mean_return(*args, **kwargs) -> float:
    """Find the mean return of a policy.

    Shortcut to call `generate_trajectories` and fetch the `rollout_stats` value for
    `'return_mean'`; see documentation for `generate_trajectories` and `rollout_stats`.
    """
    trajectories = generate_trajectories(*args, **kwargs)
    return rollout_stats(trajectories)["return_mean"]


def flatten_trajectories(
    trajectories: Sequence[types.Trajectory],
) -> types.Transitions:
    """Flatten a series of trajectory dictionaries into arrays.
    Returns observations, actions, next observations, rewards.
    Args:
        trajectories: list of trajectories.
    Returns:
      The trajectories flattened into a single batch of Transitions.
    """
    keys = ["obs", "next_obs", "acts", "dones", "infos"]
    parts = {key: [] for key in keys}
    for traj in trajectories:
        parts["acts"].append(traj.acts)

        obs = traj.obs
        parts["obs"].append(obs[:-1])
        parts["next_obs"].append(obs[1:])

        dones = np.zeros(len(traj.acts), dtype=np.bool)
        dones[-1] = True
        parts["dones"].append(dones)

        if traj.infos is None:
            infos = np.array([{}] * len(traj))
        else:
            infos = traj.infos
        parts["infos"].append(infos)

    cat_parts = {
        key: np.concatenate(part_list, axis=0) for key, part_list in parts.items()
    }
    lengths = set(map(len, cat_parts.values()))
    assert len(lengths) == 1, f"expected one length, got {lengths}"
    return types.Transitions(**cat_parts)


def flatten_trajectories_with_rew(
    trajectories: Sequence[types.TrajectoryWithRew],
) -> types.TransitionsWithRew:
    transitions = flatten_trajectories(trajectories)
    rews = np.concatenate([traj.rews for traj in trajectories])
    return types.TransitionsWithRew(**dataclasses.asdict(transitions), rews=rews)


def generate_transitions(
    policy, venv, n_timesteps: int, *, truncate: bool = True, **kwargs
) -> types.TransitionsWithRew:
    """Generate obs-action-next_obs-reward tuples.

    Args:
      policy (BasePolicy or BaseRLModel): A stable_baselines policy or RLModel,
          trained on the gym environment.
      venv: The vectorized environments to interact with.
      n_timesteps: The minimum number of timesteps to sample.
      truncate: If True, then drop any additional samples to ensure that exactly
          `n_timesteps` samples are returned.
      **kwargs: Passed-through to generate_trajectories.

    Returns:
      A batch of Transitions. The length of the constituent arrays is guaranteed
      to be at least `n_timesteps` (if specified), but may be greater unless
      `truncate` is provided as we collect data until the end of each episode.
    """
    traj = generate_trajectories(
        policy, venv, sample_until=min_timesteps(n_timesteps), **kwargs
    )
    transitions = flatten_trajectories_with_rew(traj)
    if truncate and n_timesteps is not None:
        as_dict = dataclasses.asdict(transitions)
        truncated = {k: arr[:n_timesteps] for k, arr in as_dict.items()}
        transitions = types.TransitionsWithRew(**truncated)
    return transitions


def rollout_and_save(
    path: str,
    policy: Union[BaseRLModel, BasePolicy],
    venv: VecEnv,
    sample_until: GenTrajTerminationFn,
    *,
    unwrap: bool = True,
    exclude_infos: bool = True,
    verbose: bool = True,
    **kwargs,
) -> None:
    """Generate policy rollouts and save them to a pickled list of trajectories.

    The `.infos` field of each Trajectory is set to `None` to save space.

    Args:
      path: Rollouts are saved to this path.
      venv: The vectorized environments.
      sample_until: End condition for rollout sampling.
      unwrap: If True, then save original observations and rewards (instead of
        potentially wrapped observations and rewards) by calling
        `unwrap_traj()`.
      exclude_infos: If True, then exclude `infos` from pickle by setting
        this field to None. Excluding `infos` can save a lot of space during
        pickles.
      verbose: If True, then print out rollout stats before saving.
      deterministic_policy: Argument from `generate_trajectories`.
    """
    trajs = generate_trajectories(policy, venv, sample_until, **kwargs)
    if unwrap:
        trajs = [unwrap_traj(traj) for traj in trajs]
    if exclude_infos:
        trajs = [dataclasses.replace(traj, infos=None) for traj in trajs]
    if verbose:
        stats = rollout_stats(trajs)
        tf.logging.info(f"Rollout stats: {stats}")

    types.save(path, trajs)



"""Density-based baselines for imitation learning.

Each of these algorithms learns a density estimate on some aspect of the demonstrations,
then rewards the agent for following that estimate.
"""

from typing import Sequence

import numpy as np
import tensorflow as tf
from gym.spaces.utils import flatten
from sklearn.neighbors import KernelDensity
from sklearn.preprocessing import StandardScaler
from stable_baselines.common.base_class import BaseRLModel
from stable_baselines.common.vec_env import VecEnv

from imitation.data import rollout, types
from imitation.util import reward_wrapper

# Constants identifying different kinds of density we can use. Note that all
# can be augmented to depend on the time step by passing `is_stationary = True`
# to `DensityReward`.

# Density on state s
STATE_DENSITY = "state_density"
# Density on (s,a) pairs
STATE_ACTION_DENSITY = "state_action_density"
# Density (s,s') pairs
STATE_STATE_DENSITY = "state_state_density"


class DensityReward:
    def __init__(
        self,
        *,
        trajectories,
        density_type,
        kernel,
        kernel_bandwidth,
        obs_space,
        act_space,
        is_stationary=True,
        standardise_inputs=True,
    ):
        """Reward function based on a density estimate of trajectories.

        Args:
          trajectories ([dict]): expert demonstration trajectories generated by,
            e.g., `imitation.util.rollout.generate_trajectories`.
          obs_space (gym.Space): observation space for underlying environment.
          act_space (gym.Space): action space for underlying environment.
          density_type (str): type of density to train on. Valid choices are
            defined as constants in this module, including `STATE_DENSITY` (density
            on single states), `STATE_ACTION_DENSITY` (density on state-action
            pairs), and `STATE_STATE_DENSITY` (density on state-state transitions).
          is_stationary (bool): should this trainer use the same density model for
            all time steps (True), or a different density model for each time step
            (False)? `is_stationary=False` might be useful if using
            `density_type=STATE_DENSITY`, since it encourages the agent to imitate
            entire trajectories instead of simply occupying a few states that have
            high frequency in the demonstration dataset.
          standardise_inputs (bool): if True, then the inputs to the reward model
            will be standardised to have zero mean and unit variance over the
            demonstration trajectories. Otherwise, inputs will be passed to the
            reward model with their ordinary scale.
          kernel (str): kernel to use for density estimation. Current valid options
            include `gaussian` and `exponential`.
          kernel_bandwidth (float): bandwidth of kernel. If `standardise_inputs` is
            true and you are using a Gaussian kernel, then it probably makes sense
            to set this somewhere between 0.1 and 1.
        """
        self.density_type = density_type
        self.is_stationary = is_stationary
        self.kernel = kernel
        self.kernel_bandwidth = kernel_bandwidth
        self.standardise = standardise_inputs
        self.obs_space = obs_space
        self.act_space = act_space
        self._fit_models(trajectories)

    def _fit_models(self, trajectories):
        flat_trajs = self._preprocess_trajectories(trajectories)

        # if requested, we'll scale demonstration transitions so that they have
        # zero mean and unit variance (i.e all components are equally important)
        flattened_dataset = np.stack(sum(flat_trajs, []), axis=0)
        self._scaler = StandardScaler(
            with_mean=self.standardise, with_std=self.standardise
        )
        self._scaler.fit(flattened_dataset)

        # now fit density model
        # TODO: add absorbing state fix that I describe in __call__
        if self.is_stationary:
            # fit to all pairs, since density model is stationary
            self._density_model = self._fit_single_density(
                self._scaler.transform(flattened_dataset)
            )
        else:
            # fit separately for samples at each time step
            T = max(map(len, flat_trajs))
            traj_groups = [[]] * T
            for traj in flat_trajs:
                for t, flat_trans in enumerate(traj):
                    traj_groups[t].append(flat_trans)
            traj_groups_scaled = [
                self._scaler.transform(np.stack(step_transitions, axis=0))
                for step_transitions in traj_groups
            ]
            self._density_models = [
                self._fit_single_density(scaled_flat_trans)
                for scaled_flat_trans in traj_groups_scaled
            ]

    def _fit_single_density(self, flat_transitions):
        # This bandwidth was chosen to make sense with standardised inputs that
        # have unit variance in each component. There might be a better way to
        # choose it automatically.
        density_model = KernelDensity(
            kernel=self.kernel, bandwidth=self.kernel_bandwidth
        )
        density_model.fit(flat_transitions)
        return density_model

    def _preprocess_trajectories(self, trajectories: Sequence[types.Trajectory]):
        """Preprocess a list of trajectories into atomic units that we can learn a
        density function on. Depending on configuration, that could mean a sequence
        state/state pairs, or state/action pairs, or single states, etc.

        Args:
          trajectories: trajectories to process.

        Returns:
          flat_trajectories (np.ndarray): a corresponding list of "flattened"
            trajectories. Each element of a flattened trajectory is a
            one-dimensional ndarray representing a (state,state) pair at a
            particular time step, or (state,action) pair at that time step, or
            whatever the class configuration calls for.
        """
        flat_trajectories = []
        for traj in trajectories:
            obs_vec = traj.obs
            act_vec = traj.acts
            assert len(obs_vec) == len(act_vec) + 1
            flat_traj = []
            for step_num in range(len(traj.acts)):
                flat_trans = self._preprocess_transition(
                    obs_vec[step_num], act_vec[step_num], obs_vec[step_num + 1]
                )
                flat_traj.append(flat_trans)
            flat_trajectories.append(flat_traj)
        return flat_trajectories

    def _preprocess_transition(self, obs, act, next_obs):
        if self.density_type == STATE_DENSITY:
            return flatten(self.obs_space, obs)
        elif self.density_type == STATE_ACTION_DENSITY:
            return np.concatenate(
                [flatten(self.obs_space, obs), flatten(self.act_space, act)]
            )
        elif self.density_type == STATE_STATE_DENSITY:
            return np.concatenate(
                [flatten(self.obs_space, obs), flatten(self.obs_space, next_obs)]
            )
        else:
            raise ValueError(f"Unknown density type {self.density_type}")

    def __call__(self, obs_b, act_b, next_obs_b, steps=None):
        r"""Compute reward from given (s,a,s') transition batch.

        This handles *batches* of observations, since it's designed to work with
        VecEnvs.

        Args:
          obs_b (np.ndarray): current batch of observations.
          act_b (np.ndarray): batch of actions that agent took in response to those
            observations.
          next_obs_b (np.ndarray): batch of observations encountered after the
            agent took those actions.
          steps (np.ndarray): number of time steps elapsed in each environment.

        Returns:
          rew_array (np.ndarray): array of scalar rewards of the form `r_t(s,a,s')
            = \log \hat p_t(s,a,s')` (one for each environment), where `\log \hat
            p` is the underlying density model (and may be independent of s', a, or
            t, depending on options passed to constructor)."""
        rew_list = []
        assert len(obs_b) == len(act_b) and len(obs_b) == len(next_obs_b)
        for idx, (obs, act, next_obs) in enumerate(zip(obs_b, act_b, next_obs_b)):
            flat_trans = self._preprocess_transition(obs, act, next_obs)
            scaled_padded_trans = self._scaler.transform(flat_trans[np.newaxis])
            if self.is_stationary:
                rew = self._density_model.score(scaled_padded_trans)
            else:
                time = steps[idx]
                if time >= len(self._density_models):
                    # Can't do anything sensible here yet. Correct solution is to use
                    # hierarchical model in which we first check whether state is
                    # absorbing, then assign either constant score or a score based on
                    # density.
                    raise Exception(
                        f"Time {time} out of range (0, {len(self._density_models)}], "
                        "and I haven't implemented absorbing states etc. yet"
                    )
                else:
                    time_model = self._density_models[time]
                    rew = time_model.score(scaled_padded_trans)
            rew_list.append(rew)
        rew_array = np.asarray(rew_list, dtype="float32")
        return rew_array


class DensityTrainer:
    def __init__(
        self,
        venv: VecEnv,
        rollouts: Sequence[types.Trajectory],
        imitation_trainer: BaseRLModel,
        *,
        standardise_inputs: bool = True,
        kernel: str = "gaussian",
        kernel_bandwidth: float = 0.5,
        density_type: str = STATE_ACTION_DENSITY,
        is_stationary: bool = False,
    ):
        r"""Family of simple imitation learning baseline algorithms that apply RL to
        maximise a rough density estimate of the demonstration trajectories.
        Specifically, it constructs a non-parametric estimate of `p(s)`, `p(s,s')`,
        `p_t(s,a)`, etc. (depending on options), then rewards the imitation learner
        with `r_t(s,a,s')=\log p_t(s,a,s')` (or `\log p(s,s')`, or whatever the
        user wants the model to condition on).

        Args:
            venv: environment to train on.
            rollouts: list of expert trajectories to imitate.
            imitation_trainer: RL algorithm & initial policy that will
                be used to train the imitation learner.
            kernel, kernel_bandwidth, density_type, is_stationary,
                n_expert_trajectories: these are passed directly to `DensityReward`;
                refer to documentation for that class."""
        self.venv = venv
        self.imitation_trainer = imitation_trainer
        self.reward_fn = DensityReward(
            trajectories=rollouts,
            density_type=density_type,
            obs_space=self.venv.observation_space,
            act_space=self.venv.action_space,
            is_stationary=is_stationary,
            kernel=kernel,
            kernel_bandwidth=kernel_bandwidth,
            standardise_inputs=standardise_inputs,
        )
        self.wrapped_env = reward_wrapper.RewardVecEnvWrapper(self.venv, self.reward_fn)
        self.graph = tf.Graph()
        self.sess = tf.Session(graph=self.graph)
        with self.graph.as_default():
            self.sess.run(tf.global_variables_initializer())

    def train_policy(self, n_timesteps=int(1e6), **kwargs):
        """Train the imitation policy for a given number of timesteps.

        Args:
            n_timesteps (int): number of timesteps to train the policy for.
            kwargs (dict): extra arguments that will be passed to the `learn()`
                method of the imitation RL model. Refer to Stable Baselines docs for
                details.
        """
        self.imitation_trainer.set_env(self.wrapped_env)
        # FIXME: learn() is not meant to be called frequently; there are
        # significant per-call overheads (see Adam's comment in adversarial.py)
        # FIXME: the ep_reward_mean reported by SB is wrong; it comes from a
        # Monitor() that is being (incorrectly) used to wrap the underlying
        # environment.
        self.imitation_trainer.learn(
            n_timesteps,
            # ensure we can see total steps for all
            # learn() calls, not just for this call
            reset_num_timesteps=False,
            **kwargs,
        )

    def test_policy(self, *, n_trajectories=10, true_reward=True):
        """Test current imitation policy on environment & give some rollout stats.

        Args:
            n_trajectories (int): number of rolled-out trajectories.
            true_reward (bool): should this use ground truth reward from underlying
                environment (True), or imitation reward (False)?

        Returns:
            dict: rollout statistics collected by
                `imitation.utils.rollout.rollout_stats()`.
        """
        self.imitation_trainer.set_env(self.venv)
        trajs = rollout.generate_trajectories(
            self.imitation_trainer,
            self.venv if true_reward else self.wrapped_env,
            sample_until=rollout.min_episodes(n_trajectories),
        )
        reward_stats = rollout.rollout_stats(trajs)
        return reward_stats



"""model_builder.py A small library for programmatically building MuJoCo XML files."""
import tempfile
from contextlib import contextmanager

import numpy as np


def default_model(name):
    """Get a model with basic settings such as gravity and RK4 integration."""
    model = MJCModel(name)
    root = model.root

    # Setup
    root.compiler(angle="radian", inertiafromgeom="true")
    default = root.default()
    default.joint(armature=1, damping=1, limited="true")
    default.geom(contype=0, friction="1 0.1 0.1", rgba="0.7 0.7 0 1")
    root.option(gravity="0 0 -9.81", integrator="RK4", timestep=0.01)
    return model


def pointmass_model(name):
    """Get a model with basic settings such as gravity and Euler integration."""
    model = MJCModel(name)
    root = model.root

    # Setup
    root.compiler(angle="radian", inertiafromgeom="true", coordinate="local")
    default = root.default()
    default.joint(limited="false", damping=1)
    default.geom(
        contype=2,
        conaffinity="1",
        condim="1",
        friction=".5 .1 .1",
        density="1000",
        margin="0.002",
    )
    root.option(timestep=0.01, gravity="0 0 0", iterations="20", integrator="Euler")
    return model


class MJCModel(object):
    def __init__(self, name):
        self.name = name
        self.root = MJCTreeNode("mujoco").add_attr("model", name)

    @contextmanager
    def asfile(self):
        """Usage:
        model = MJCModel('reacher')
        with model.asfile() as f:
            print f.read()  # prints a dump of the model
        """
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".xml", delete=True) as f:
            self.root.write(f)
            f.seek(0)
            yield f

    def find_attr(self, attr, value):
        return self.root.find_attr(attr, value)

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        pass


class MJCTreeNode(object):
    def __init__(self, name):
        self.name = name
        self.attrs = {}
        self.children = []

    def add_attr(self, key, value):
        if isinstance(value, str):
            pass
        elif isinstance(value, list) or isinstance(value, np.ndarray):
            value = " ".join([str(val).lower() for val in value])
        else:
            value = str(value).lower()

        self.attrs[key] = value
        return self

    def __getattr__(self, name):
        def wrapper(**kwargs):
            newnode = MJCTreeNode(name)
            for (k, v) in kwargs.items():
                newnode.add_attr(k, v)
            self.children.append(newnode)
            return newnode

        return wrapper

    def dfs(self):
        yield self
        if self.children:
            for child in self.children:
                for node in child.dfs():
                    yield node

    def find_attr(self, attr, value):
        """Run DFS to find a matching attr."""
        if attr in self.attrs and self.attrs[attr] == value:
            return self
        for child in self.children:
            res = child.find_attr(attr, value)
            if res is not None:
                return res
        return None

    def write(self, ostream, tabs=0):
        contents = " ".join(['%s="%s"' % (k, v) for (k, v) in self.attrs.items()])
        if self.children:
            ostream.write("\t" * tabs)
            ostream.write("<%s %s>\n" % (self.name, contents))
            for child in self.children:
                child.write(ostream, tabs=tabs + 1)
            ostream.write("\t" * tabs)
            ostream.write("</%s>\n" % self.name)
        else:
            ostream.write("\t" * tabs)
            ostream.write("<%s %s/>\n" % (self.name, contents))

    def __str__(self):
        s = "<" + self.name
        s += " ".join(['%s="%s"' % (k, v) for (k, v) in self.attrs.items()])
        return s + ">"



import json
import os
import warnings
from typing import Any, Callable, List, NamedTuple, Union

import sacred


class SacredDicts(NamedTuple):
    """Each dict `foo` is loaded from `f"{sacred_dir}/foo.json"`."""

    sacred_dir: str
    config: dict
    run: dict

    @classmethod
    def load_from_dir(cls, sacred_dir: str):
        args = []
        for field in cls._fields:
            if field == "sacred_dir":
                args.append(sacred_dir)
            else:
                json_path = os.path.join(sacred_dir, f"{field}.json")
                with open(json_path, "r") as f:
                    args.append(json.load(f))
        return cls(*args)


def dir_contains_sacred_jsons(dir_path: str) -> bool:
    run_path = os.path.join(dir_path, "run.json")
    config_path = os.path.join(dir_path, "config.json")
    return os.path.isfile(run_path) and os.path.isfile(config_path)


def filter_subdirs(
    root_dir: str,
    filter_fn: Callable[[str], bool] = dir_contains_sacred_jsons,
    *,
    nested_ok: bool = False,
) -> List[str]:
    """Walks through a directory tree, returning paths to filtered subdirectories.

    Does not follow symlinks.

    Args:
      root_dir: The start of the directory tree walk.
      filter_fn: A function with takes a directory path and returns True if
        we should include the directory path in this function's return value.
      nested_ok: If False, then error if in the return value, one of the
        directory paths is a subdirectory of another.

    Returns:
      A list of all subdirectory paths where `filter_fn(path) == True`.
    """
    filtered_dirs = set()
    for root, _, _ in os.walk(root_dir):
        if filter_fn(root):
            filtered_dirs.add(root)

    if not nested_ok:
        for dirpath in filtered_dirs:
            components = os.path.split(dirpath)
            for i in range(1, len(components)):
                prefix = os.path.join(*components[0:i])
                if prefix in filtered_dirs:
                    raise ValueError(f"Parent {prefix} to {dir} also a dir directory")
    return list(filtered_dirs)


def build_sacred_symlink(log_dir: str, run: sacred.run.Run) -> None:
    """Constructs a symlink "{log_dir}/sacred" => "${SACRED_PATH}"."""
    sacred_dir = get_sacred_dir_from_run(run)
    if sacred_dir is None:
        warnings.warn(RuntimeWarning("Couldn't find sacred directory."))
        return
    symlink_path = os.path.join(log_dir, "sacred")
    target_path = os.path.relpath(sacred_dir, start=log_dir)
    # Use relative paths so we can mount the output directory at different paths
    # (e.g. when copying across machines).
    os.symlink(target_path, symlink_path, target_is_directory=True)


def get_sacred_dir_from_run(run: sacred.run.Run) -> Union[str, None]:
    """Returns path to the sacred directory, or None if not found."""
    for obs in run.observers:
        if isinstance(obs, sacred.observers.FileStorageObserver):
            return obs.dir
    return None


def dict_get_nested(d: dict, nested_key: str, *, sep=".", default=None) -> Any:
    curr = d
    for key in nested_key.split(sep):
        if isinstance(curr, dict) and key in curr:
            curr = curr[key]
        else:
            return default
    return curr



import collections.abc
import copy
import os
from typing import Any, Callable, Dict, Optional

import ray
import ray.tune
import sacred
from sacred.observers import FileStorageObserver

from imitation.scripts.config.parallel import parallel_ex


@parallel_ex.main
def parallel(
    sacred_ex_name: str,
    run_name: str,
    search_space: dict,
    base_named_configs: list,
    base_config_updates: dict,
    resources_per_trial: dict,
    init_kwargs: Dict[str, Any],
    local_dir: Optional[str],
    upload_dir: Optional[str],
) -> None:
    """Parallelize multiple runs of another Sacred Experiment using Ray Tune.

    A Sacred FileObserver is attached to the inner experiment and writes Sacred
    logs to "{RAY_LOCAL_DIR}/sacred/". These files are automatically copied over
    to `upload_dir` if that argument is provided.

    Args:
        sacred_ex_name: The Sacred experiment to tune. Either "expert_demos" or
            "train_adversarial".
        run_name: A name describing this parallelizing experiment.
            This argument is also passed to `ray.tune.run` as the `name` argument.
            It is also saved in 'sacred/run.json' of each inner Sacred experiment
            under the 'experiment.name' key. This is equivalent to using the Sacred
            CLI '--name' option on the inner experiment. Offline analysis jobs can use
            this argument to group similar data.
        search_space: A dictionary which can contain Ray Tune search objects like
            `ray.tune.grid_search` and `ray.tune.sample_from`, and is
            passed as the `config` argument to `ray.tune.run()`. After the
            `search_space` is transformed by Ray, it passed into
            `sacred_ex.run(**run_kwargs)` as `run_kwargs` (`sacred_ex` is the Sacred
            Experiment selected via `sacred_ex_name`). Usually `search_space` only has
            the keys "named_configs" and "config_updates", but any parameter names
            to `sacred.Experiment.run()` are okay.
        base_named_configs: Default Sacred named configs. Any named configs
            taken from `search_space` are higher priority than the base_named_configs.
            Concretely, this priority is implemented by appending named configs taken
            from `search_space` to the run's named configs after `base_named_configs`.

            Named configs in `base_named_configs` don't appear in the automatically
            generated Ray directory name, unlike named configs from `search_space`.

        base_config_updates: Default Sacred config updates. Any config updates taken
            from `search_space` are higher priority than `base_config_updates`.

            Config updates in `base_config_updates` don't appear in the automatically
            generated Ray directory name, unlike config updates from `search_space`.
        resources_per_trial: Argument to `ray.tune.run()`.
        init_kwargs: Arguments to pass to `ray.init`.
        local_dir: `local_dir` argument to `ray.tune.run()`.
        upload_dir: `upload_dir` argument to `ray.tune.run()`.
    """
    # Basic validation for config options before we enter parallel jobs.
    for name in base_named_configs:
        assert isinstance(name, str)
    for k in base_config_updates:
        assert isinstance(k, str)
    assert isinstance(search_space["named_configs"], collections.abc.Sequence)
    assert isinstance(search_space["config_updates"], collections.abc.Mapping)

    # Convert Sacred's ReadOnlyList to List because not picklable.
    base_named_configs = list(base_named_configs)

    # Convert Sacred's ReadOnlyDict (and recursively convert ReadOnlyContainer values)
    # to regular python variants because not picklable.
    base_config_updates = copy.deepcopy(base_config_updates)
    search_space = copy.deepcopy(search_space)

    # Explicitly set `data_dir` if parallelizing `train_adversarial`. We need this to
    # automatically find rollout pickles because Ray sets a new working directory for
    # each Raylet.
    if sacred_ex_name == "train_adversarial":
        if "data_dir" not in base_config_updates:
            data_dir = os.path.join(os.getcwd(), "data/")
            base_config_updates["data_dir"] = data_dir

    trainable = _ray_tune_sacred_wrapper(
        sacred_ex_name,
        run_name,
        base_named_configs,
        base_config_updates,
    )

    # Disable all Ray Loggers.
    #
    # JSON and CSV loggers are redundant now that we have Sacred logs.
    # TensorBoard logs don't contain useful information (inner Sacred experiment
    # never gets access to `reporter`), and clog up the TensorBoard Runs
    # dashboard.
    ray_loggers = ()

    ray.init(**init_kwargs)
    try:
        ray.tune.run(
            trainable,
            config=search_space,
            name=run_name,
            local_dir=local_dir,
            upload_dir=upload_dir,
            loggers=ray_loggers,
            resources_per_trial=resources_per_trial,
        )
    finally:
        ray.shutdown()


def _ray_tune_sacred_wrapper(
    sacred_ex_name: str,
    run_name: str,
    base_named_configs: list,
    base_config_updates: dict,
) -> Callable:
    """From an Experiment build a wrapped run function suitable for Ray Tune.

    `ray.tune.run(...)` expects a trainable function that takes a dict
    argument `config`. The wrapped function uses `config` as keyword args for
    `ex.run(...)` because we want to be able to hyperparameter tune over both the
    `named_configs` and `config_updates` arguments.

    The Ray Tune `reporter` is not passed to the inner experiment.

    Args have the same meanings as arguments described in `parallel`.

    Returns:
      A function that takes two arguments, `config` (used as keyword args for
      `ex.run`) and `reporter`. The function returns the run result.
    """

    def inner(config: dict, reporter) -> dict:
        """Trainable function with the correct signature for `ray.tune`.

        Args:
            config: Keyword arguments for `ex.run()`, where `ex` is the
                `sacred.Experiment` instance associated with `sacred_ex_name`.
        """
        # Set Sacred capture mode to "sys" because default "fd" option leads to error.
        # See https://github.com/IDSIA/sacred/issues/289.
        # TODO(shwang): Stop modifying CAPTURE_MODE once the issue is fixed.
        sacred.SETTINGS.CAPTURE_MODE = "sys"

        run_kwargs = config
        updated_run_kwargs = {}
        # Import inside function rather than in module because Sacred experiments
        # are not picklable, and Ray requires this function to be picklable.
        from imitation.scripts.expert_demos import expert_demos_ex
        from imitation.scripts.train_adversarial import train_ex

        experiments = {
            "expert_demos": expert_demos_ex,
            "train_adversarial": train_ex,
        }
        ex = experiments[sacred_ex_name]

        observer = FileStorageObserver("sacred")
        ex.observers.append(observer)

        # Apply base configs to get modified `named_configs` and `config_updates`.
        named_configs = []
        named_configs.extend(base_named_configs)
        named_configs.extend(run_kwargs["named_configs"])
        updated_run_kwargs["named_configs"] = named_configs

        config_updates = {}
        config_updates.update(base_config_updates)
        config_updates.update(run_kwargs["config_updates"])
        updated_run_kwargs["config_updates"] = config_updates

        # Add other run_kwargs items to updated_run_kwargs.
        for k, v in run_kwargs.items():
            if k not in updated_run_kwargs:
                updated_run_kwargs[k] = v

        run = ex.run(**updated_run_kwargs, options={"--run": run_name})

        # Ray Tune has a string formatting error if raylet completes without
        # any calls to `reporter`.
        reporter(done=True)

        assert run.status == "COMPLETED"
        return run.result

    return inner


def main_console():
    observer = FileStorageObserver(os.path.join("output", "sacred", "parallel"))
    parallel_ex.observers.append(observer)
    parallel_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()
